#!/usr/bin/python

#Benchmark for Distortion correction in PyFAI

from __future__ import print_function, division

import json, sys, time, timeit, os, platform, subprocess, gc, logging
logging.basicConfig(level=logging.ERROR)
import numpy
import fabio

import os.path as op
sys.path.append(op.join(op.dirname(op.dirname(op.abspath(__file__))), "test"))
import utilstest

try:
    from rfoo.utils import rconsole
    rconsole.spawn_server()
except ImportError:
    print("No socket opened for debugging -> please install rfoo")

#We use the locally build version of PyFAI
pyFAI = utilstest.UtilsTest.pyFAI
ocl = pyFAI.opencl.ocl
import matplotlib
matplotlib.use("gtk")
from matplotlib import pyplot as plt
plt.ion()

splinefile = utilstest.UtilsTest.getimage("1900/frelon.spline")

detector = pyFAI.detectors.FReLoN(splinefile)

repeat = 1
number = 1

print("Number of iteration: %s average over %s processing" % (repeat, number))
for method in ("LUT","CSR"):
    for device in ("None", '"CPU"', '"GPU"'):
        for wg in [1, 2, 4, 8, 16, 32]:
            if (method != "CSR" or device == "None")and wg > 1:
                continue
            setup = """
import pyFAI, pyFAI.distortion, numpy
detector = pyFAI.detectors.FReLoN("%s")
dis = pyFAI.distortion.Distortion(detector, method='%s', device=%s, workgroup=%s)
data = numpy.random.randint(0,65000,size=detector.shape[0]*detector.shape[1]).reshape(detector.shape).astype(numpy.uint16)
dis.calc_init()""" % (splinefile, method, device, wg)
            t = timeit.Timer("dis.correct(data)", setup)
            tmin = min([i / number for i in t.repeat(repeat=repeat, number=number)])
            print("%s %s (wg=%2s) t=%.3fms" % (method, device, wg, tmin * 1000.0))
            gc.collect



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#    Project: Fast Azimuthal Integration
#             https://github.com/pyFAI/pyFAI
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"test suite for convolution cython code"

__author__ = "Jérôme Kieffer"
__contact__ = "Jérôme Kieffer"
__license__ = "GPLv3+"
__copyright__ = "European Synchrotron Radiation Facility, Grenoble, France"
__date__ = "15/12/2014"

import sys
import unittest
import numpy
if __name__ == '__main__':
    import pkgutil, os
    __path__ = pkgutil.extend_path([os.path.dirname(__file__)], "pyFAI.test")
from .utilstest import getLogger  # UtilsTest, Rwp, getLogger
logger = getLogger(__file__)
pyFAI = sys.modules["pyFAI"]
from pyFAI import _convolution
import scipy.ndimage, scipy.misc, scipy.signal

class TestConvolution(unittest.TestCase):
    def setUp(self):
        self.sigma = 1
        self.width = 8 * self.sigma + 1
        if self.width % 2 == 0:
            self.width += 1
        self.gauss = scipy.signal.gaussian(self.width, self.sigma)
        self.gauss /= self.gauss.sum()
        self.lena = scipy.misc.lena().astype("float32")

    def test_gaussian(self):
        gauss = _convolution.gaussian(self.sigma)
        self.assert_(numpy.allclose(gauss, self.gauss), "gaussian curves are the same")

    def test_horizontal_convolution(self):
        gauss = self.gauss.astype(numpy.float32)
        ref = scipy.ndimage.filters.convolve1d(self.lena, self.gauss, axis=-1)
        obt = _convolution.horizontal_convolution(self.lena, gauss)
        self.assert_(numpy.allclose(ref, obt), "horizontal filtered images are the same")

    def test_vertical_convolution(self):
        gauss = self.gauss.astype(numpy.float32)
        ref = scipy.ndimage.filters.convolve1d(self.lena, self.gauss, axis=0)
        obt = _convolution.vertical_convolution(self.lena, gauss)
        self.assert_(numpy.allclose(ref, obt), "vertical filtered images are the same")

    def test_gaussian_filter(self):
        ref = scipy.ndimage.filters.gaussian_filter(self.lena, self.sigma)
        obt = _convolution.gaussian_filter(self.lena, self.sigma)
        self.assert_(numpy.allclose(ref, obt), "gaussian filtered images are the same")


def test_suite_all_convolution():
    testSuite = unittest.TestSuite()
    testSuite.addTest(TestConvolution("test_horizontal_convolution"))
    testSuite.addTest(TestConvolution("test_vertical_convolution"))
    testSuite.addTest(TestConvolution("test_gaussian"))
    testSuite.addTest(TestConvolution("test_gaussian_filter"))
    return testSuite

if __name__ == '__main__':
    mysuite = test_suite_all_convolution()
    runner = unittest.TextTestRunner()
    runner.run(mysuite)



#!/usr/bin/python
import pyFAI, numpy
img = numpy.zeros((512,512))
for i in range(1,6):img[i*100,i*100]=1
det = pyFAI.detectors.Detector(1e-4,1e-4)
det.shape=(512,512)
ai=pyFAI.AzimuthalIntegrator(1,detector=det)
import pylab
from utilstest import Rwp
results = {}
for i, meth in enumerate(["cython", "splitbbox", "splitpixel", "csr_no", "csr_bbox", "csr_full"]):
    tth, I = ai.integrate1d(img, 10000, method=meth, unit="2th_deg")
    pylab.plot(tth, I + i * 1e-3, label=meth)
    ai.reset()
    results[meth]=tth, I
print("no_split R=%.3f" % Rwp(results["csr_no"], results["cython"]))
print("split_bbox R=%.3f" % Rwp(results["csr_bbox"], results["splitbbox"]))
print("split_full R=%.3f" % Rwp(results["csr_full"], results["splitpixel"]))
pylab.legend()
pylab.ion()
pylab.show()
raw_input("enter_to_quit")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#    Project: Fast Azimuthal Integration
#             https://github.com/pyFAI/pyFAI
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"test suite for Distortion correction class"

__author__ = "Jérôme Kieffer"
__contact__ = "Jerome.Kieffer@ESRF.eu"
__license__ = "GPLv3+"
__copyright__ = "European Synchrotron Radiation Facility, Grenoble, France"
__date__ = "15/12/2014"


import unittest
# import os
import numpy
# import logging, time
import sys
import fabio
if __name__ == '__main__':
    import pkgutil, os
    __path__ = pkgutil.extend_path([os.path.dirname(__file__)], "pyFAI.test")
from .utilstest import UtilsTest, getLogger
logger = getLogger(__file__)
pyFAI = sys.modules["pyFAI"]
from pyFAI import _distortion, detectors
# _distortion = sys.modules["pyFAI._distortion"]
# detectors = sys.modules["pyFAI.detectors"]

class test_halfccd(unittest.TestCase):
    """basic test"""
    halfFrelon = "1464/LaB6_0020.edf"
    splineFile = "1461/halfccd.spline"
    fit2d_cor = "2454/halfccd.fit2d.edf"

    def setUp(self):
        """Download files"""
        self.fit2dFile = UtilsTest.getimage(self.__class__.fit2d_cor)
        self.halfFrelon = UtilsTest.getimage(self.__class__.halfFrelon)
        self.splineFile = UtilsTest.getimage(self.__class__.splineFile)
        self.det = detectors.FReLoN(self.splineFile)
        self.dis = _distortion.Distortion(self.det)
        self.fit2d = fabio.open(self.fit2dFile).data
        self.raw = fabio.open(self.halfFrelon).data

    def test_vs_fit2d(self):
        """
        Compare spline correction vs fit2d's code

        precision at 1e-3 : 90% of pixels
        """
        size = self.dis.calc_LUT_size()
        mem = size.max() * self.raw.nbytes * 4 / 2.0 ** 20
        logger.info("Memory expected for LUT: %.3f MBytes", mem)
        try:
            self.dis.calc_LUT()
        except MemoryError as error:
            logger.warning("test_halfccd.test_vs_fit2d failed because of MemoryError. This test tries to allocate %.3fMBytes and failed with %s", mem, error)
            return
        cor = self.dis.correct(self.raw)
        delta = abs(cor - self.fit2d)
        mask = numpy.where(self.fit2d == 0)
        denom = self.fit2d.copy()
        denom[mask] = 1
        ratio = delta / denom
        ratio[mask] = 0
        good_points_ratio = 1.0 * (ratio < 1e-3).sum() / self.raw.size
        logger.info("ratio of good points (less than 1/1000 relative error): %.4f" % good_points_ratio)
        self.assert_(good_points_ratio > 0.99, "99% of all points have a relative error below 1/1000")

def test_suite_all_distortion():
    testSuite = unittest.TestSuite()
    testSuite.addTest(test_halfccd("test_vs_fit2d"))
#    testSuite.addTest(test_azim_halfFrelon("test_numpy_vs_fit2d"))
#    testSuite.addTest(test_azim_halfFrelon("test_cythonSP_vs_fit2d"))
#    testSuite.addTest(test_azim_halfFrelon("test_cython_vs_numpy"))
#    testSuite.addTest(test_flatimage("test_splitPixel"))
#    testSuite.addTest(test_flatimage("test_splitBBox"))
# This test is known to be broken ...
#    testSuite.addTest(test_saxs("test_mask"))

    return testSuite

if __name__ == '__main__':

    mysuite = test_suite_all_distortion()
    runner = unittest.TextTestRunner()
    runner.run(mysuite)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#    Project: Fast Azimuthal Integration 
#             https://github.com/pyFAI/pyFAI
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
# 
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
# 
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__doc__ = """
Calculate bounding box around   
"""
__author__ = "Jerome Kieffer"
__contact__ = "Jerome.kieffer@esrf.fr"
__date__ = "20/10/2014"
__status__ = "stable"
__license__ = "GPLv3+"
import cython
cimport numpy
import numpy

ctypedef numpy.int64_t DTYPE_int64_t
ctypedef numpy.float64_t DTYPE_float64_t


def boundingBox(data):
    """
    Calculate bounding box around   

    @param img: 2D array like
    @return: 4-typle (d0_min, d1_min, d0_max, d1_max)
    
    
    NOTA: Does not work :( 
     
    """
    cdef numpy.ndarray[long, ndim = 1] shape = numpy.array(data.shape)
    cdef long ndims = data.ndim

    cdef numpy.ndarray[long, ndim = 1] mins = numpy.array(shape)
    cdef numpy.ndarray[long, ndim = 1] maxs = numpy.zeros(ndims, dtype=int)

    cdef long  i = 0
    cdef long  j = 0
    cdef long  k = 0
    cdef long  l = 0
    if ndims == 1:
            for i in range(shape[0]):
                if data[i] > 0.0:
                    if i < mins[0]:
                        mins[0] = i
                    if i > maxs[0]:
                        maxs[0] = i
    elif ndims == 2:
            for i in range(shape[0]):
                for j in range(shape[1]):
                    if data[i, j] > 0.0 :
                        if i < mins[0]:
                            mins[0] = i
                        if i > maxs[0]:
                            maxs[0] = i
                        if j < mins[1]:
                            mins[1] = i
                        if j > maxs[1]:
                            maxs[1] = i
    elif ndims == 3:
            for i in range(shape[0]):
                for j in range(shape[1]):
                    for k in range(shape[2]):
                        if  data[i, j, k] > 0.0:
                            if i < mins[0]:
                                mins[0] = i
                            if i > maxs[0]:
                                maxs[0] = i
                            if j < mins[1]:
                                mins[1] = i
                            if j > maxs[1]:
                                maxs[1] = i
                            if k < mins[2]:
                                mins[2] = i
                            if k > maxs[2]:
                                maxs[2] = i
    elif ndims == 4:
            for i in range(shape[0]):
                for j in range(shape[1]):
                    for k in range(shape[2]):
                        for l in range(shape[3]):
                            if data[i, j, k, l] > 0.0:
                                if i < mins[0]:
                                    mins[0] = i
                                if i > maxs[0]:
                                    maxs[0] = i
                                if j < mins[1]:
                                    mins[1] = i
                                if j > maxs[1]:
                                    maxs[1] = i
                                if k < mins[2]:
                                    mins[2] = i
                                if k > maxs[2]:
                                    maxs[2] = i
                                if l < mins[3]:
                                    mins[3] = i
                                if l > maxs[3]:
                                    maxs[3] = i
    else:
        raise RuntimeError("Dimensions > 4 not implemented")
    return tuple(mins) + tuple(maxs)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#    Project: Fast Azimuthal Integration
#             https://github.com/pyFAI/pyFAI
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__doc__ = """
Full pixel Splitting implemented using Sparse-matrix Dense-Vector multiplication,
Sparse matrix represented using the CompressedSparseROw.
"""
__author__ = "Jerome Kieffer"
__contact__ = "Jerome.kieffer@esrf.fr"
__date__ = "08/04/2015"
__status__ = "stable"
__license__ = "GPLv3+"

import cython
import os
import sys
from cython.parallel import prange
from libc.string cimport memset
import numpy
cimport numpy
from libc.math cimport fabs, floor
from libc.stdio cimport printf

include "regrid_common.pxi"

try:
    from fastcrc import crc32
except:
    from zlib import crc32

cdef struct Function:
    double slope
    double intersect


cdef double area4(double a0, double a1, double b0, double b1, double c0, double c1, double d0, double d1) nogil:
    """
    Calculate the area of the ABCD quadrilataire  with corners:
    A(a0,a1)
    B(b0,b1)
    C(c0,c1)
    D(d0,d1)
    @return: area, i.e. 1/2 * (AC ^ BD)
    """
    return 0.5 * fabs(((c0 - a0) * (d1 - b1)) - ((c1 - a1) * (d0 - b0)))


cdef double integrate(double A0, double B0, Function AB) nogil:
    """
    integrates the line defined by AB, from A0 to B0
    param A0: first limit
    param B0: second limit
    param AB: struct with the slope and point of intersection of the line
    """
    if A0 == B0:
        return 0.0
    else:
        return AB.slope * (B0 * B0 - A0 * A0) * 0.5 + AB.intersect * (B0 - A0)


class HistoLUT1dFullSplit(object):
    """
    Now uses CSR (Compressed Sparse raw) with main attributes:
    * nnz: number of non zero elements
    * data: coefficient of the matrix in a 1D vector of float64
    * indices: Column index position for the data (same size as
    * indptr: row pointer indicates the start of a given row. len nrow+1

    Nota: nnz = indptr[-1]
    """
    @cython.boundscheck(False)
    def __init__(self,
                 numpy.ndarray pos not None,
                 int bins=100,
                 pos0Range=None,
                 pos1Range=None,
                 mask=None,
                 mask_checksum=None,
                 allow_pos0_neg=False,
                 unit="undefined",
                 bad_pixel=False):
        """
        @param pos: 3D or 4D array with the coordinates of each pixel point
        @param bins: number of output bins, 100 by default
        @param pos0Range: minimum and maximum  of the 2th range
        @param pos1Range: minimum and maximum  of the chi range
        @param mask: array (of int8) with masked pixels with 1 (0=not masked)
        @param allow_pos0_neg: enforce the q<0 is usually not possible
        @param unit: can be 2th_deg or r_nm^-1 ...
        """

#        self.padding = int(padding)
        if pos.ndim>3: #create a view
            pos = pos.reshape((-1,4,2))
        assert pos.shape[1] == 4
        assert pos.shape[2] == 2
        assert pos.ndim == 3
        self.pos = pos
        self.size = pos.shape[0]
        self.bins = bins
        self.bad_pixel = bad_pixel
        self.lut_size = 0
        self.allow_pos0_neg = allow_pos0_neg
        if  mask is not None:
            assert mask.size == self.size
            self.check_mask = True
            self.cmask = numpy.ascontiguousarray(mask.ravel(), dtype=numpy.int8)
            if mask_checksum:
                self.mask_checksum = mask_checksum
            else:
                self.mask_checksum = crc32(mask)
        else:
            self.check_mask = False
            self.mask_checksum = None
        self.data = self.nnz = self.indices = self.indptr = None
        self.pos0Range = pos0Range
        self.pos1Range = pos1Range

        self.calc_lut()
        self.outPos = numpy.linspace(self.pos0_min+0.5*self.delta, self.pos0_maxin-0.5*self.delta, self.bins)
        self.lut_checksum = crc32(self.data)
        self.unit=unit
        self.lut=(self.data,self.indices,self.indptr)
        self.lut_nbytes = sum([i.nbytes for i in self.lut])

    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    def calc_lut(self):
        cdef:
            numpy.ndarray[numpy.float64_t, ndim = 3] cpos = numpy.ascontiguousarray(self.pos,dtype=numpy.float64)
            numpy.int8_t[:] cmask
            numpy.ndarray[numpy.int32_t, ndim = 1] outMax = numpy.zeros(self.bins, dtype=numpy.int32)
            numpy.ndarray[numpy.int32_t, ndim = 1] indptr = numpy.zeros(self.bins+1, dtype=numpy.int32)
            double pos0_min=0, pos0_max=0, pos0_maxin=0, pos1_min=0, pos1_max=0, pos1_maxin=0
            double max0, min0
            double areaPixel=0, delta=0, areaPixel2=0
            double A0=0, B0=0, C0=0, D0=0, A1=0, B1=0, C1=0, D1=0
            double A_lim=0, B_lim=0, C_lim=0, D_lim=0
            double oneOverArea=0, partialArea=0, tmp=0
            Function AB, BC, CD, DA
            int bins, i=0, idx=0, bin=0, bin0_max=0, bin0_min=0, pixel_bins=0, k=0, size=0
            bint check_pos1=False, check_mask=False
            int range1=0, range2=0

        bins = self.bins
        if self.pos0Range is not None and len(self.pos0Range) > 1:
            self.pos0_min = min(self.pos0Range)
            self.pos0_maxin = max(self.pos0Range)
        else:
            self.pos0_min = self.pos[:, :, 0].min()
            self.pos0_maxin = self.pos[:, :, 0].max()
        self.pos0_max = self.pos0_maxin * (1 + numpy.finfo(numpy.float64).eps)
        if self.pos1Range is not None and len(self.pos1Range) > 1:
            self.pos1_min = min(self.pos1Range)
            self.pos1_maxin = max(self.pos1Range)
            self.check_pos1 = True
        else:
            self.pos1_min = self.pos[:, :, 1].min()
            self.pos1_maxin = self.pos[:, :, 1].max()
        self.pos1_max = self.pos1_maxin * (1 + numpy.finfo(numpy.float64).eps)

        self.delta = (self.pos0_max - self.pos0_min) / (< double > (bins))

        pos0_min = self.pos0_min
        pos0_max = self.pos0_max
        pos1_min = self.pos1_min
        pos1_max = self.pos1_max
        delta = self.delta

        size = self.size
        check_mask = self.check_mask
        if check_mask:
            cmask = self.cmask

        if self.bad_pixel:
            range1 = self.bad_pixel
            range2 = self.bad_pixel + 1
        else:
            range1 = 0
            range2 = size

        print( "FLOAT 64")
        print( "++++++++")
        print( "Space bounds: %e  -  %e" % (pos0_min, pos0_max))

        with nogil:
            for idx in range(range1,range2):

#                 with gil:
#                     print "Pixel %d" % idx
#                     print "==========="
#                     print "==========="

                if (check_mask) and (cmask[idx]):
                    continue

                A0 = get_bin_number(< double > cpos[idx, 0, 0], pos0_min, delta)
                A1 = < double > cpos[idx, 0, 1]
                B0 = get_bin_number(< double > cpos[idx, 1, 0], pos0_min, delta)
                B1 = < double > cpos[idx, 1, 1]
                C0 = get_bin_number(< double > cpos[idx, 2, 0], pos0_min, delta)
                C1 = < double > cpos[idx, 2, 1]
                D0 = get_bin_number(< double > cpos[idx, 3, 0], pos0_min, delta)
                D1 = < double > cpos[idx, 3, 1]

#                 with gil:
#                     print "Pixel in 2-th"
#                     print "============="
#                     print "A: %e --> %e  %e" % (< double > cpos[idx, 0, 0], A0, A1)
#                     print "B: %e --> %e  %e" % (< double > cpos[idx, 1, 0], B0, B1)
#                     print "C: %e --> %e  %e" % (< double > cpos[idx, 2, 0], C0, C1)
#                     print "D: %e --> %e  %e" % (< double > cpos[idx, 3, 0], D0, D1)
#                     print " "
#


                min0 = min(A0, B0, C0, D0)
                max0 = max(A0, B0, C0, D0)

#                 with gil:
#                     print "Min 2-th: %e" % min0
#                     print "Max 2-th: %e" % max0
#                     print " "

                if (max0<0) or (min0 >=bins):
                    continue
                if check_pos1:
                    if (max(A1, B1, C1, D1) < pos1_min) or (min(A1, B1, C1, D1) > pos1_maxin):
                        continue

                bin0_min = < int > floor(min0)
                bin0_max = < int > floor(max0)

#                 with gil:
#                     print "Bin span: %d - %d" % (bin0_min, bin0_max)
#                     print " "


                for bin in range(bin0_min, bin0_max+1):
                    outMax[bin] += 1

        indptr[1:] = outMax.cumsum()
        self.indptr = indptr

        cdef:
            numpy.ndarray[numpy.int32_t, ndim = 1] indices = numpy.zeros(indptr[bins], dtype=numpy.int32)
            numpy.ndarray[numpy.float64_t, ndim = 1] data = numpy.zeros(indptr[bins], dtype=numpy.float64)
            numpy.ndarray[numpy.float32_t, ndim = 1] areas = numpy.zeros(indptr[bins], dtype=numpy.float32)

        #just recycle the outMax array
        memset(&outMax[0], 0, bins * sizeof(numpy.int32_t))

        with nogil:
            for idx in range(range1,range2):

                if (check_mask) and (cmask[idx]):
                    continue

                A0 = get_bin_number(< double > cpos[idx, 0, 0], pos0_min, delta)
                A1 = < double > cpos[idx, 0, 1]
                B0 = get_bin_number(< double > cpos[idx, 1, 0], pos0_min, delta)
                B1 = < double > cpos[idx, 1, 1]
                C0 = get_bin_number(< double > cpos[idx, 2, 0], pos0_min, delta)
                C1 = < double > cpos[idx, 2, 1]
                D0 = get_bin_number(< double > cpos[idx, 3, 0], pos0_min, delta)
                D1 = < double > cpos[idx, 3, 1]

#                 with gil:
#                     print "...and again..."
#                     print "Pixel in 2-th"
#                     print "============="
#                     print "A: %e --> %e  %e" % (< double > cpos[idx, 0, 0], A0, A1)
#                     print "B: %e --> %e  %e" % (< double > cpos[idx, 1, 0], B0, B1)
#                     print "C: %e --> %e  %e" % (< double > cpos[idx, 2, 0], C0, C1)
#                     print "D: %e --> %e  %e" % (< double > cpos[idx, 3, 0], D0, D1)
#                     print " "




                min0 = min(A0, B0, C0, D0)
                max0 = max(A0, B0, C0, D0)

#                 with gil:
#                     print "Min 2-th: %e" % min0
#                     print "Max 2-th: %e" % max0
#                     print " "
#
                if (max0<0) or (min0 >=bins):
                    continue
                if check_pos1:
                    if (max(A1, B1, C1, D1) < pos1_min) or (min(A1, B1, C1, D1) > pos1_maxin):
                        continue

                bin0_min = < int > floor(min0)
                bin0_max = < int > floor(max0)

#                 with gil:
#                     print "Bin span: %d - %d" % (bin0_min, bin0_max)
#                     print " "


                if bin0_min == bin0_max:
                    #All pixel is within a single bin
                    k = outMax[bin0_min]
                    indices[indptr[bin0_min]+k] = idx
                    data[indptr[bin0_min]+k] = 1.0
                    areas[indptr[bin0_min]+k] = -1.0
                    outMax[bin0_min] += 1 #k+1
                else:  #else we have pixel spliting.z
                    AB.slope=(B1-A1)/(B0-A0)
                    AB.intersect= A1 - AB.slope*A0
                    BC.slope=(C1-B1)/(C0-B0)
                    BC.intersect= B1 - BC.slope*B0
                    CD.slope=(D1-C1)/(D0-C0)
                    CD.intersect= C1 - CD.slope*C0
                    DA.slope=(A1-D1)/(A0-D0)
                    DA.intersect= D1 - DA.slope*D0

#                     with gil:
#                         print "The lines that make up the pixel:"
#                         print "================================="
#                         print "AB: %e  %e" % (AB.slope, AB.intersect)
#                         print "BC: %e  %e" % (BC.slope, BC.intersect)
#                         print "CD: %e  %e" % (CD.slope, CD.intersect)
#                         print "DA: %e  %e" % (DA.slope, DA.intersect)
#                         print " "

                    areaPixel = area4(A0, A1, B0, B1, C0, C1, D0, D1)

#                     with gil:
#                         print "Area with the 4 point formula: %e" % areaPixel
#                         print " "

                    areaPixel2  = integrate(A0, B0, AB)
                    areaPixel2 += integrate(B0, C0, BC)
                    areaPixel2 += integrate(C0, D0, CD)
                    areaPixel2 += integrate(D0, A0, DA)

#                     with gil:
#                         print "(Area with integration: %e)" % areaPixel2
#                         print " "

                    oneOverPixelArea = 1.0 / areaPixel

#                     with gil:
#                         print "1 over area: %e" % oneOverPixelArea
#                         print " "

                    partialArea2 = 0.0
                    for bin in range(bin0_min, bin0_max+1):
#                         with gil:
#                             print "Bin: %d" % bin
#                             print "======="
#                             print "  "
                        A_lim = (A0<=bin)*(A0<=(bin+1))*bin + (A0>bin)*(A0<=(bin+1))*A0 + (A0>bin)*(A0>(bin+1))*(bin+1)
                        B_lim = (B0<=bin)*(B0<=(bin+1))*bin + (B0>bin)*(B0<=(bin+1))*B0 + (B0>bin)*(B0>(bin+1))*(bin+1)
                        C_lim = (C0<=bin)*(C0<=(bin+1))*bin + (C0>bin)*(C0<=(bin+1))*C0 + (C0>bin)*(C0>(bin+1))*(bin+1)
                        D_lim = (D0<=bin)*(D0<=(bin+1))*bin + (D0>bin)*(D0<=(bin+1))*D0 + (D0>bin)*(D0>(bin+1))*(bin+1)


#                         with gil:
#                             print "Limits:"
#                             print "======="
#                             print "A_lim = %d*bin + %d*A0 + %d*(bin+1) = %e" % ((A0<=bin)*(A0<=(bin+1)),(A0>bin)*(A0<=(bin+1)),(A0>bin)*(A0>(bin+1)),A_lim)
#                             print "B_lim = %d*bin + %d*B0 + %d*(bin+1) = %e" % ((B0<=bin)*(B0<=(bin+1)),(B0>bin)*(B0<=(bin+1)),(B0>bin)*(B0>(bin+1)),B_lim)
#                             print "C_lim = %d*bin + %d*C0 + %d*(bin+1) = %e" % ((C0<=bin)*(C0<=(bin+1)),(C0>bin)*(C0<=(bin+1)),(C0>bin)*(C0>(bin+1)),C_lim)
#                             print "D_lim = %d*bin + %d*D0 + %d*(bin+1) = %e" % ((D0<=bin)*(D0<=(bin+1)),(D0>bin)*(D0<=(bin+1)),(D0>bin)*(D0>(bin+1)),D_lim)
#                             print "  "
#
                        partialArea  = integrate(A_lim, B_lim, AB)
                        partialArea += integrate(B_lim, C_lim, BC)
                        partialArea += integrate(C_lim, D_lim, CD)
                        partialArea += integrate(D_lim, A_lim, DA)
                        tmp = fabs(partialArea) * oneOverPixelArea
#                         with gil:
#                             print "Partial Area: %e" % fabs(partialArea)
#                             print "Contribution: %e" % tmp
#                             print "  "
                        k = outMax[bin]
                        indices[indptr[bin]+k] = idx
                        data[indptr[bin]+k] = tmp
                        outMax[bin] += 1 #k+1
        self.data = data
        self.indices = indices
        self.outMax = outMax
        self.areas = areas

    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    def integrate(self, weights, dummy=None, delta_dummy=None, dark=None, flat=None, solidAngle=None, polarization=None):
        """
        Actually perform the integration which in this case looks more like a matrix-vector product

        @param weights: input image
        @type weights: ndarray
        @param dummy: value for dead pixels (optional)
        @type dummy: float
        @param delta_dummy: precision for dead-pixel value in dynamic masking
        @type delta_dummy: float
        @param dark: array with the dark-current value to be subtracted (if any)
        @type dark: ndarray
        @param flat: array with the dark-current value to be divided by (if any)
        @type flat: ndarray
        @param solidAngle: array with the solid angle of each pixel to be divided by (if any)
        @type solidAngle: ndarray
        @param polarization: array with the polarization correction values to be divided by (if any)
        @type polarization: ndarray
        @return : positions, pattern, weighted_histogram and unweighted_histogram
        @rtype: 4-tuple of ndarrays
        """
        cdef:
            numpy.int32_t i=0, j=0, idx=0, bins=self.bins, size=self.size
            double sum_data=0.0, sum_count=0.0, epsilon=1e-10
            double data=0, coef=0, cdummy=0, cddummy=0
            bint do_dummy=False, do_dark=False, do_flat=False, do_polarization=False, do_solidAngle=False
            numpy.ndarray[numpy.float64_t, ndim = 1] outData = numpy.zeros(self.bins, dtype=numpy.float64)
            numpy.ndarray[numpy.float64_t, ndim = 1] outCount = numpy.zeros(self.bins, dtype=numpy.float64)
            numpy.ndarray[numpy.float64_t, ndim = 1] outMerge = numpy.zeros(self.bins, dtype=numpy.float64)
            double[:] ccoef = self.data, cdata, tdata, cflat, cdark, csolidAngle, cpolarization
            numpy.int32_t[:] indices = self.indices, indptr = self.indptr
        assert size == weights.size

        if dummy is not None:
            do_dummy = True
            cdummy =  <double>float(dummy)
            if delta_dummy is None:
                cddummy = <double>0.0
            else:
                cddummy = <double>float(delta_dummy)

        if flat is not None:
            do_flat = True
            assert flat.size == size
            cflat = numpy.ascontiguousarray(flat.ravel(), dtype=numpy.float64)
        if dark is not None:
            do_dark = True
            assert dark.size == size
            cdark = numpy.ascontiguousarray(dark.ravel(), dtype=numpy.float64)
        if solidAngle is not None:
            do_solidAngle = True
            assert solidAngle.size == size
            csolidAngle = numpy.ascontiguousarray(solidAngle.ravel(), dtype=numpy.float64)
        if polarization is not None:
            do_polarization = True
            assert polarization.size == size
            cpolarization = numpy.ascontiguousarray(polarization.ravel(), dtype=numpy.float64)

        if (do_dark + do_flat + do_polarization + do_solidAngle):
            tdata = numpy.ascontiguousarray(weights.ravel(), dtype=numpy.float64)
            cdata = numpy.zeros(size,dtype=numpy.float64)
            if do_dummy:
                for i in prange(size, nogil=True, schedule="static"):
                    data = tdata[i]
                    if ((cddummy!=0) and (fabs(data-cdummy) > cddummy)) or ((cddummy==0) and (data!=cdummy)):
                        # Nota: -= and /= operatore are seen as reduction in cython parallel.
                        if do_dark:
                            data = data - cdark[i]
                        if do_flat:
                            data = data / cflat[i]
                        if do_polarization:
                            data = data / cpolarization[i]
                        if do_solidAngle:
                            data = data / csolidAngle[i]
                        cdata[i]+=data
                    else:
                        # set all dummy_like values to cdummy. simplifies further processing
                        cdata[i]+=cdummy
            else:
                for i in prange(size, nogil=True, schedule="static"):
                    data = tdata[i]
                    if do_dark:
                        data = data - cdark[i]
                    if do_flat:
                        data = data / cflat[i]
                    if do_polarization:
                        data = data / cpolarization[i]
                    if do_solidAngle:
                        data = data / csolidAngle[i]
                    cdata[i] += data
        else:
            if do_dummy:
                tdata = numpy.ascontiguousarray(weights.ravel(), dtype=numpy.float64)
                cdata = numpy.zeros(size,dtype=numpy.float64)
                for i in prange(size, nogil=True, schedule="static"):
                    data = tdata[i]
                    if ((cddummy != 0) and (fabs(data - cdummy) > cddummy)) or ((cddummy == 0) and (data != cdummy)):
                        cdata[i] += data
                    else:
                        cdata[i] += cdummy
            else:
                cdata = numpy.ascontiguousarray(weights.ravel(), dtype=numpy.float64)

        for i in prange(bins, nogil=True, schedule="guided"):
            sum_data = 0.0
            sum_count = 0.0
            for j in range(indptr[i], indptr[i + 1]):
                idx = indices[j]
                coef = ccoef[j]
                if coef == 0.0:
                    continue
                data = cdata[idx]
                if do_dummy and data == cdummy:
                    continue
                sum_data = sum_data + coef * data
                sum_count = sum_count + coef
            outData[i] += sum_data
            outCount[i] += sum_count
            if sum_count > epsilon:
                outMerge[i] += sum_data / sum_count
            else:
                outMerge[i] += cdummy
        return self.outPos, outMerge, outData, outCount





#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#    Project: Fast Azimuthal Integration
#             https://github.com/pyFAI/pyFAI
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"test suite for OpenCL code"

__author__ = "Jérôme Kieffer"
__contact__ = "Jerome.Kieffer@ESRF.eu"
__license__ = "GPLv3+"
__copyright__ = "European Synchrotron Radiation Facility, Grenoble, France"
__date__ = "02/04/2015"


import unittest
import os
import time
import sys
import fabio
import gc
import tempfile
import numpy
import platform
if __name__ == '__main__':
    import pkgutil
    __path__ = pkgutil.extend_path([os.path.dirname(__file__)], "pyFAI.test")
from .utilstest import UtilsTest, Rwp, getLogger, recursive_delete
logger = getLogger(__file__)
try:
    import pyopencl
except ImportError as error:
    logger.warning("OpenCL module (pyopencl) is not present, skip tests. %s." % error)
    skip = True
else:
    skip = False

pyFAI = sys.modules["pyFAI"]
from pyFAI.opencl import ocl
if ocl is None:
    skip = True
else:
    pyopencl = pyFAI.opencl.pyopencl
    import pyopencl.array


class TestMask(unittest.TestCase):
    tmp_dir = tempfile.mkdtemp(prefix="pyFAI_test_OpenCL_")
    N = 1000

    def setUp(self):
        self.datasets = [{"img": UtilsTest.getimage("1883/Pilatus1M.edf"),
                          "poni": UtilsTest.getimage("1893/Pilatus1M.poni"),
                          "spline": None},
                         {"img": UtilsTest.getimage("1882/halfccd.edf"),
                          "poni": UtilsTest.getimage("1895/halfccd.poni"),
                          "spline": UtilsTest.getimage("1461/halfccd.spline")},
                         {"img": UtilsTest.getimage("1881/Frelon2k.edf"),
                          "poni": UtilsTest.getimage("1896/Frelon2k.poni"),
                          "spline": UtilsTest.getimage("1900/frelon.spline")},
                         {"img": UtilsTest.getimage("1884/Pilatus6M.cbf"),
                          "poni": UtilsTest.getimage("1897/Pilatus6M.poni"),
                          "spline": None},
            ]
        if not os.path.isdir(self.tmp_dir):
            os.makedirs(self.tmp_dir)
        for ds in self.datasets:
            if ds["spline"] is not None:
                data = open(ds["poni"], "r").read()
#                spline = os.path.basename(ds["spline"])
                with open(ds["poni"]) as f:
                    data = []
                    for line in f:
                        if line.startswith("SplineFile:"):
                            data.append("SplineFile: " + ds["spline"])
                        else:
                            data.append(line.strip())
                ds["poni"] = os.path.join(self.tmp_dir, os.path.basename(ds["poni"]))
                with open(ds["poni"], "w") as f:
                    f.write(os.linesep.join(data))

    def tearDown(self):
        recursive_delete(self.tmp_dir)

    def test_OpenCL(self):
        logger.info("Testing histogram-based algorithm (forward-integration)")
        for devtype in ("GPU", "CPU"):
            ids = ocl.select_device(devtype, extensions=["cl_khr_int64_base_atomics"])
            if ids is None:
                logger.error("No suitable %s OpenCL device found" % devtype)
                continue
            else:
                logger.info("I found a suitable device %s %s: %s %s " % (devtype, ids, ocl.platforms[ids[0]], ocl.platforms[ids[0]].devices[ids[1]]))

            for ds in self.datasets:
                ai = pyFAI.load(ds["poni"])
                data = fabio.open(ds["img"]).data
                res = ai.xrpd_OpenCL(data, self.N, devicetype="all", platformid=ids[0], deviceid=ids[1], useFp64=True)
                ref = ai.integrate1d(data, self.N, method="splitBBox", unit="2th_deg")
                r = Rwp(ref, res)
                logger.info("OpenCL histogram vs histogram SplitBBox has R= %.3f for dataset %s" % (r, ds))
                self.assertTrue(r < 6, "Rwp=%.3f for OpenCL histogram processing of %s" % (r, ds))
                del ai, data
                gc.collect()

    def test_OpenCL_LUT(self):
        logger.info("Testing LUT-based algorithm (backward-integration)")
        for devtype in ("GPU", "CPU"):
            ids = ocl.select_device(devtype, best=True)
            if ids is None:
                logger.error("No suitable %s OpenCL device found" % devtype)
                continue
            else:
                logger.info("I found a suitable device %s %s: %s %s " % (devtype, ids, ocl.platforms[ids[0]], ocl.platforms[ids[0]].devices[ids[1]]))

            for ds in self.datasets:
                ai = pyFAI.load(ds["poni"])
                data = fabio.open(ds["img"]).data
                ref = ai.integrate1d(data, self.N, method="splitBBox", unit="2th_deg")
                try:
                    res = ai.integrate1d(data, self.N, method="ocl_lut_%i,%i" % (ids[0], ids[1]), unit="2th_deg")
                except (pyFAI.opencl.pyopencl.MemoryError, MemoryError, pyFAI.opencl.pyopencl.RuntimeError, RuntimeError) as error:
                    logger.warning("Memory error on %s dataset %s: %s%s. Converted into warnining: device may not have enough memory." % (devtype, os.path.basename(ds["img"]), os.linesep, error))
                    break
                else:
                    ref = ai.xrpd(data, self.N)
                    r = Rwp(ref, res)
                    logger.info("OpenCL CSR vs histogram SplitBBox has R= %.3f for dataset %s" % (r, ds))
                    self.assertTrue(r < 3, "Rwp=%.3f for OpenCL LUT processing of %s" % (r, ds))
                del ai, data
                gc.collect()

    def test_OpenCL_CSR(self):
        logger.info("Testing CSR-based algorithm (backward-integration)")
        for devtype in ("GPU", "CPU"):
            ids = ocl.select_device(devtype, best=True)
            if ids is None:
                logger.error("No suitable %s OpenCL device found" % devtype)
                continue
            else:
                logger.info("I found a suitable device %s %s: %s %s " % (devtype, ids, ocl.platforms[ids[0]], ocl.platforms[ids[0]].devices[ids[1]]))

            for ds in self.datasets:
                ai = pyFAI.load(ds["poni"])
                data = fabio.open(ds["img"]).data
                ref = ai.integrate1d(data, self.N, method="splitBBox", unit="2th_deg")
                try:
                    res = ai.integrate1d(data, self.N, method="ocl_csr_%i,%i" % (ids[0], ids[1]), unit="2th_deg")
                except (pyFAI.opencl.pyopencl.MemoryError, MemoryError, pyFAI.opencl.pyopencl.RuntimeError, RuntimeError) as error:
                    logger.warning("Memory error on %s dataset %s: %s%s. Converted into Warning: device may not have enough memory." % (devtype, os.path.basename(ds["img"]), os.linesep, error))
                    break
                else:
                    r = Rwp(ref, res)
                    logger.info("OpenCL CSR vs histogram SplitBBox has R= %.3f for dataset %s" % (r, ds))
                    self.assertTrue(r < 3, "Rwp=%.3f for OpenCL CSR processing of %s" % (r, ds))
                del ai, data
                gc.collect()


class TestSort(unittest.TestCase):
    """
    Test the kernels for vector and image sorting
    """
    N = 1024
    ws = N // 8

    def setUp(self):
        self.h_data = numpy.random.random(self.N).astype("float32")
        self.h2_data = numpy.random.random((self.N, self.N)).astype("float32").reshape((self.N, self.N))

        self.ctx = ocl.create_context(devicetype="GPU")
        device = self.ctx.devices[0]
        try:
            devtype = pyopencl.device_type.to_string(device.type).upper()
        except ValueError:
            # pocl does not describe itself as a CPU !
            devtype = "CPU"
        workgroup = device.max_work_group_size
        if (devtype == "CPU") and (device.platform.vendor == "Apple"):
            logger.info("For Apple's OpenCL on CPU: enforce max_work_goup_size=1")
            workgroup = 1

        self.ws = min(workgroup, self.ws)
        self.queue = pyopencl.CommandQueue(self.ctx, properties=pyopencl.command_queue_properties.PROFILING_ENABLE)
        self.local_mem = pyopencl.LocalMemory(self.ws * 32)  # 2float4 = 2*4*4 bytes per workgroup size
        src = pyFAI.utils.read_cl_file("bitonic.cl")
        self.prg = pyopencl.Program(self.ctx, src).build()

    def tearDown(self):
        self.h_data = None
        self.queue = None
        self.ctx = None
        self.local_mem = None
        self.h2_data = None

    def test_reference_book(self):
        d_data = pyopencl.array.to_device(self.queue, self.h_data)
        t0 = time.time()
        hs_data = numpy.sort(self.h_data)
        t1 = time.time()
        time_sort = 1e3 * (t1 - t0)

        evt = self.prg.bsort_book(self.queue, (self.ws,), (self.ws,), d_data.data, self.local_mem)
        evt.wait()
        err = abs(hs_data - d_data.get()).max()
        logger.info("test_reference_book")
        logger.info("Numpy sort on %s element took %s ms" % (self.N, time_sort))
        logger.info("Reference sort time: %s ms, err=%s " % (1e-6 * (evt.profile.end - evt.profile.start), err))
        # this test works under linux:
        if platform.system() == "Linux":
            self.assert_(err == 0.0)
        else:
            logger.warning("Measured error on %s is %s" % (platform.system(), err))

    def test_reference_file(self):
        d_data = pyopencl.array.to_device(self.queue, self.h_data)
        t0 = time.time()
        hs_data = numpy.sort(self.h_data)
        t1 = time.time()
        time_sort = 1e3 * (t1 - t0)

        evt = self.prg.bsort_file(self.queue, (self.ws,), (self.ws,), d_data.data, self.local_mem)
        evt.wait()
        err = abs(hs_data - d_data.get()).max()
        logger.info("test_reference_file")
        logger.info("Numpy sort on %s element took %s ms" % (self.N, time_sort))
        logger.info("Reference sort time: %s ms, err=%s " % (1e-6 * (evt.profile.end - evt.profile.start), err))
        # this test works anywhere !
        self.assert_(err == 0.0)

    def test_sort_all(self):
        d_data = pyopencl.array.to_device(self.queue, self.h_data)
        t0 = time.time()
        hs_data = numpy.sort(self.h_data)
        t1 = time.time()
        time_sort = 1e3 * (t1 - t0)

        evt = self.prg.bsort_all(self.queue, (self.ws,), (self.ws,), d_data.data, self.local_mem)
        evt.wait()
        err = abs(hs_data - d_data.get()).max()
        logger.info("test_sort_all")
        logger.info("Numpy sort on %s element took %s ms" % (self.N, time_sort))
        logger.info("modified function execution time: %s ms, err=%s " % (1e-6 * (evt.profile.end - evt.profile.start), err))
        self.assert_(err == 0.0)

    def test_sort_horizontal(self):
        d2_data = pyopencl.array.to_device(self.queue, self.h2_data)
        t0 = time.time()
        h2s_data = numpy.sort(self.h2_data, axis=-1)
        t1 = time.time()
        time_sort = 1e3 * (t1 - t0)
        evt = self.prg.bsort_horizontal(self.queue, (self.N, self.ws), (1, self.ws), d2_data.data, self.local_mem)
        evt.wait()
        err = abs(h2s_data - d2_data.get()).max()
        logger.info("Numpy horizontal sort on %sx%s elements took %s ms" % (self.N, self.N, time_sort))
        logger.info("Horizontal execution time: %s ms, err=%s " % (1e-6 * (evt.profile.end - evt.profile.start), err))
        self.assert_(err == 0.0)

    def test_sort_vertical(self):
        d2_data = pyopencl.array.to_device(self.queue, self.h2_data)
        t0 = time.time()
        h2s_data = numpy.sort(self.h2_data, axis=0)
        t1 = time.time()
        time_sort = 1e3 * (t1 - t0)
        evt = self.prg.bsort_vertical(self.queue, (self.ws, self.N), (self.ws, 1), d2_data.data, self.local_mem)
        evt.wait()
        err = abs(h2s_data - d2_data.get()).max()
        logger.info("Numpy vertical sort on %sx%s elements took %s ms" % (self.N, self.N, time_sort))
        logger.info("Vertical execution time: %s ms, err=%s " % (1e-6 * (evt.profile.end - evt.profile.start), err))
        self.assert_(err == 0.0)


def test_suite_all_OpenCL():
    testSuite = unittest.TestSuite()
    if skip:
        logger.warning("OpenCL module (pyopencl) is not present or no device available: skip tests")
    else:
        testSuite.addTest(TestMask("test_OpenCL"))
        testSuite.addTest(TestMask("test_OpenCL_LUT"))
        testSuite.addTest(TestMask("test_OpenCL_CSR"))
        testSuite.addTest(TestSort("test_reference_book"))
        testSuite.addTest(TestSort("test_reference_file"))
        testSuite.addTest(TestSort("test_sort_all"))
        testSuite.addTest(TestSort("test_sort_horizontal"))
        testSuite.addTest(TestSort("test_sort_vertical"))
    return testSuite

if __name__ == '__main__':

    mysuite = test_suite_all_OpenCL()
    runner = unittest.TextTestRunner()
    runner.run(mysuite)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#    Project: Fast Azimuthal Integration
#             https://github.com/pyFAI/pyFAI
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

__doc__ = """
Calculates histograms of pos0 (tth) weighted by Intensity

Splitting is done on the pixel's bounding box like fit2D,
reverse implementation based on a sparse matrix multiplication
"""
__author__ = "Jerome Kieffer"
__contact__ = "Jerome.kieffer@esrf.fr"
__date__ = "08/05/2015"
__status__ = "stable"
__license__ = "GPLv3+"

import cython
import os
import sys
from cpython.ref cimport PyObject, Py_XDECREF
from cython.parallel import prange
from libc.string cimport memset, memcpy
from cython cimport view
import numpy
cimport numpy
from libc.math cimport fabs
from libcpp.vector cimport vector

cdef struct lut_point:
    numpy.int32_t idx
    numpy.float32_t coef
dtype_lut = numpy.dtype([("idx" ,numpy.int32), ("coef", numpy.float32)])
try:
    from fastcrc import crc32
except:
    from zlib import crc32
cdef bint NEED_DECREF = sys.version_info<(2,7) and numpy.version.version<"1.5"

include "regrid_common.pxi"


class HistoBBox1d(object):
    """
    1D histogramming with pixel splitting based on a Look-up table

    The initialization of the class can take quite a while (operation are not parallelized)
    but each integrate is parallelized and quite efficient.
    """

    @cython.boundscheck(False)
    def __init__(self,
                 pos0,
                 delta_pos0,
                 pos1=None,
                 delta_pos1=None,
                 int bins=100,
                 pos0Range=None,
                 pos1Range=None,
                 mask=None,
                 mask_checksum=None,
                 allow_pos0_neg=False,
                 unit="undefined"):
        """
        @param pos0: 1D array with pos0: tth or q_vect or r ...
        @param delta_pos0: 1D array with delta pos0: max center-corner distance
        @param pos1: 1D array with pos1: chi
        @param delta_pos1: 1D array with max pos1: max center-corner distance, unused !
        @param bins: number of output bins, 100 by default
        @param pos0Range: minimum and maximum  of the 2th range
        @param pos1Range: minimum and maximum  of the chi range
        @param mask: array (of int8) with masked pixels with 1 (0=not masked)
        @param allow_pos0_neg: enforce the q<0 is usually not possible
        @param unit: can be 2th_deg or r_nm^-1 ...
        """

        self.size = pos0.size
        assert delta_pos0.size == self.size
        self.bins = bins
        self.lut_size = 0
        self.allow_pos0_neg =  allow_pos0_neg
        if  mask is not None:
            assert mask.size == self.size
            self.check_mask = True
            self.cmask = numpy.ascontiguousarray(mask.ravel(), dtype=numpy.int8)
            if mask_checksum:
                self.mask_checksum = mask_checksum
            else:
                self.mask_checksum = crc32(mask)
        else:
            self.check_mask = False
            self.mask_checksum = None

        self.cpos0 = numpy.ascontiguousarray(pos0.ravel(), dtype=numpy.float32)
        self.dpos0 = numpy.ascontiguousarray(delta_pos0.ravel(), dtype=numpy.float32)
        self.cpos0_sup = numpy.empty_like(self.cpos0) #self.cpos0 + self.dpos0
        self.cpos0_inf = numpy.empty_like(self.cpos0) #self.cpos0 - self.dpos0
        self.pos0Range = pos0Range
        self.pos1Range = pos1Range
        self.calc_boundaries(pos0Range)
        if pos1Range is not None and len(pos1Range) > 1:
            assert pos1.size == self.size
            assert delta_pos1.size == self.size
            self.check_pos1 = True
            self.cpos1_min = numpy.ascontiguousarray((pos1-delta_pos1).ravel(), dtype=numpy.float32)
            self.cpos1_max = numpy.ascontiguousarray((pos1+delta_pos1).ravel(), dtype=numpy.float32)
            self.pos1_min = min(pos1Range)
            pos1_maxin = max(pos1Range)
            self.pos1_max = pos1_maxin * EPS32
        else:
            self.check_pos1 = False
            self.cpos1_min = None
            self.pos1_max = None

        self.delta = (self.pos0_max - self.pos0_min) / bins
        self._lut=None
        self.lut_max_idx = None
        self.calc_lut()
        self.outPos = numpy.linspace(self.pos0_min+0.5*self.delta, self.pos0_maxin-0.5*self.delta, self.bins)
        self.lut_checksum = None
        self.unit=unit
        self.csr_data = None
        self.csr_indices = None
        self.csr_indptr = None
#        self.generate_csr()
 #       self.generate_csr_padded()
#        self.generate_lut_csr()


    @cython.boundscheck(False)
    @cython.wraparound(False)
    def calc_boundaries(self,pos0Range):
        """
        Called by constructor to calculate the boundaries and the bin position
        """
        cdef numpy.int32_t size = self.cpos0.size
        cdef bint check_mask = self.check_mask
        cdef numpy.int8_t[:] cmask
        cdef float[:] cpos0, dpos0, cpos0_sup, cpos0_inf,
        cdef float upper, lower, pos0_max, pos0_min, c, d
        cdef bint allow_pos0_neg=self.allow_pos0_neg

        cpos0_sup = self.cpos0_sup
        cpos0_inf = self.cpos0_inf
        cpos0 = self.cpos0
        dpos0 = self.dpos0
        pos0_min=cpos0[0]
        pos0_max=cpos0[0]

        if check_mask:
            cmask = self.cmask
        with nogil:
            for idx in range(size):
                c = cpos0[idx]
                d = dpos0[idx]
                lower = c - d
                upper = c + d
                cpos0_sup[idx] = upper
                cpos0_inf[idx] = lower
                if not allow_pos0_neg and lower<0:
                    lower=0
                if not (check_mask and cmask[idx]):
                    if upper>pos0_max:
                        pos0_max = upper
                    if lower<pos0_min:
                        pos0_min = lower

        if pos0Range is not None and len(pos0Range) > 1:
            self.pos0_min = min(pos0Range)
            self.pos0_maxin = max(pos0Range)
        else:
            self.pos0_min = pos0_min
            self.pos0_maxin = pos0_max
        if (not allow_pos0_neg) and self.pos0_min < 0:
            self.pos0_min = 0
        self.pos0_max = self.pos0_maxin * EPS32


    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    def calc_lut(self):
        'calculate the max number of elements in the LUT and populate it'
        cdef float delta=self.delta, pos0_min=self.pos0_min, pos1_min, pos1_max, min0, max0, fbin0_min, fbin0_max, deltaL, deltaR, deltaA
        cdef numpy.int32_t k,idx, bin0_min, bin0_max, bins = self.bins, lut_size, i, size
        cdef bint check_mask, check_pos1
        cdef numpy.ndarray[numpy.int32_t, ndim = 1] outMax = numpy.zeros(bins, dtype=numpy.int32)
        cdef float[:] cpos0_sup = self.cpos0_sup
        cdef float[:] cpos0_inf = self.cpos0_inf
        cdef float[:] cpos1_min, cpos1_max
#        cdef numpy.float32_t[:,:] lutcoef
#        cdef int[:,:] lutidx
        cdef vector[vector[lut_point]] lut
        lut.resize(bins)
        cdef lut_point tmp
        #cdef lut_point[:,:] lut
        cdef numpy.int8_t[:] cmask
        size = self.size
        if self.check_mask:
            cmask = self.cmask
            check_mask = True
        else:
            check_mask = False

        if self.check_pos1:
            check_pos1 = True
            cpos1_min = self.cpos1_min
            cpos1_max = self.cpos1_max
            pos1_max = self.pos1_max
            pos1_min = self.pos1_min
        else:
            check_pos1 = False
#NOGIL
        with nogil:
            for idx in range(size):
                if (check_mask) and (cmask[idx]):
                    continue

                min0 = cpos0_inf[idx]
                max0 = cpos0_sup[idx]

                if check_pos1 and ((cpos1_max[idx] < pos1_min) or (cpos1_min[idx] > pos1_max)):
                        continue

                fbin0_min = get_bin_number(min0, pos0_min, delta)
                fbin0_max = get_bin_number(max0, pos0_min, delta)
                bin0_min = < numpy.int32_t > fbin0_min
                bin0_max = < numpy.int32_t > fbin0_max

                if (bin0_max < 0) or (bin0_min >= bins):
                    continue
                if bin0_max >= bins :
                    bin0_max = bins - 1
                if  bin0_min < 0:
                    bin0_min = 0

                if bin0_min == bin0_max:
                    #All pixel is within a single bin
                    k = outMax[bin0_min]
                    tmp.idx = idx
                    tmp.coef = 1.0
                    lut[bin0_min].push_back(tmp)
                    outMax[bin0_min] += 1
                else: #we have pixel splitting.
                    deltaA = 1.0 / (fbin0_max - fbin0_min)

                    deltaL = (bin0_min + 1) - fbin0_min
                    deltaR = fbin0_max - (bin0_max)

                    k = outMax[bin0_min]
                    tmp.idx = idx
                    tmp.coef = deltaA * deltaL
                    lut[bin0_min].push_back(tmp)
                    outMax[bin0_min] += 1

                    k = outMax[bin0_max]
                    tmp.idx = idx
                    tmp.coef = deltaA * deltaR
                    lut[bin0_max].push_back(tmp)
                    outMax[bin0_max] += 1

                    if bin0_min + 1 < bin0_max:
                        for i in range(bin0_min + 1, bin0_max):
                            k = outMax[i]
                            tmp.idx = idx
                            tmp.coef = deltaA
                            lut[i].push_back(tmp)
                            outMax[i] += 1

        self.lut_max_idx = outMax
        self._lut = lut

    def get_lut(self):
        """Getter for the LUT as actual numpy array"""
        cdef int rc_before, rc_after
        rc_before = sys.getrefcount(self._lut)
        cdef lut_point[:,:] lut = self._lut
        rc_after = sys.getrefcount(self._lut)
        cdef bint need_decref = NEED_DECREF and ((rc_after-rc_before)>=2)
        cdef numpy.ndarray[numpy.float64_t, ndim=2] tmp_ary = numpy.empty(shape=self._lut.shape, dtype=numpy.float64)
        memcpy(&tmp_ary[0,0], &lut[0,0], self._lut.nbytes)
        self.lut_checksum = crc32(tmp_ary)

        #Ugly against bug#89
        if need_decref and (sys.getrefcount(self._lut)>=rc_before+2):
            print("Decref needed")
            Py_XDECREF(<PyObject *> self._lut)

#        return tmp_ary.view(dtype=dtype_lut)
        return numpy.core.records.array(tmp_ary.view(dtype=dtype_lut),
                                        shape=self._lut.shape,dtype=dtype_lut,
                                        copy=True)

    lut = property(get_lut)


    #@cython.boundscheck(True)  # For testing
    #def generate_csr(self):
        #cdef int nnz = self.lut_max_idx.sum()
        #cdef lut_point tmp
        #cdef int i, j, running_index

        #data = numpy.empty(shape=nnz, dtype=numpy.float32)
        #indices = numpy.empty(shape=nnz, dtype=numpy.int32)
        #indptr = numpy.empty(shape=self.bins+1, dtype=numpy.int32)

        #running_index = 0
        #for i in range(self.bins):
            #indptr[i] = running_index
            #for j in range(self.lut_max_idx[i]):
                #tmp = self._lut[i, j]
                #data[running_index] = tmp.coef
                #indices[running_index] = tmp.idx
                #running_index += 1
        #indptr[-1] = nnz
        #self.csr_data = data
        #self.csr_indices = indices
        #self.csr_indptr = indptr


    #@cython.boundscheck(True)  # For testing
    #def generate_csr_padded(self, workgroup_size):
        #cdef int nnz=0
        #cdef lut_point tmp
        #cdef int i, j, running_index


        #lut_max_idx_padded = numpy.empty(shape=self.bins, dtype=numpy.int32)
        #for i in range(self.bins):
            #lut_max_idx_padded[i] = (self.lut_max_idx[i] + workgroup_size - 1) & ~(workgroup_size - 1)
        #nnz = lut_max_idx_padded.sum()

        #data = numpy.empty(shape=nnz, dtype=numpy.float32)
        #indices = numpy.empty(shape=nnz, dtype=numpy.int32)
        #indptr = numpy.empty(shape=self.bins+1, dtype=numpy.int32)

        #running_index = 0
        #for i in range(self.bins):
            #indptr[i] = running_index
            #for j in range(lut_max_idx_padded[i]):
                #if j < self.lut_max_idx[i]:
                    #tmp = self._lut[i, j]
                    #data[running_index] = tmp.coef
                    #indices[running_index] = tmp.idx
                    #running_index += 1
                #else:
                    #data[running_index] = 0
                    #indices[running_index] = tmp.idx
                    #running_index += 1
        #indptr[-1] = nnz
        #self.csr_padded_data = data
        #self.csr_padded_indices = indices
        #self.csr_padded_indptr = indptr
        #max_width = lut_max_idx_padded.max()
        #return max_width



    #def generate_lut_csr(self):
        #cdef int nnz = self.bins * self.lut_max_idx.max()
        #cdef lut_point tmp
        #cdef int i, j, running_index

        #data = numpy.empty(shape=nnz, dtype=numpy.float32)
        #indices = numpy.empty(shape=nnz, dtype=numpy.int32)
        #indptr = numpy.empty(shape=self.bins+1, dtype=numpy.int32)

        #running_index = 0
        #for i in range(self.bins):
            #indptr[i] = running_index
            #for j in range(self.lut_max_idx.max()):
                #tmp = self._lut[i, j]
                #data[running_index] = tmp.coef
                #indices[running_index] = tmp.idx
                #running_index += 1
        #indptr[-1] = nnz
        #self.lut_csr_data = data
        #self.lut_csr_indices = indices
        #self.lut_csr_indptr = indptr




    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    def integrate(self, weights, dummy=None, delta_dummy=None, dark=None, flat=None, solidAngle=None, polarization=None):
        """
        Actually perform the integration which in this case looks more like a matrix-vector product

        @param weights: input image
        @type weights: ndarray
        @param dummy: value for dead pixels (optional)
        @type dummy: float
        @param delta_dummy: precision for dead-pixel value in dynamic masking
        @type delta_dummy: float
        @param dark: array with the dark-current value to be subtracted (if any)
        @type dark: ndarray
        @param flat: array with the dark-current value to be divided by (if any)
        @type flat: ndarray
        @param solidAngle: array with the solid angle of each pixel to be divided by (if any)
        @type solidAngle: ndarray
        @param polarization: array with the polarization correction values to be divided by (if any)
        @type polarization: ndarray
        @return : positions, pattern, weighted_histogram and unweighted_histogram
        @rtype: 4-tuple of ndarrays

        """
        cdef numpy.int32_t i=0, j=0, idx=0, bins=self.bins, lut_size=self.lut_size, size=self.size
        cdef double sum_data=0, sum_count=0, epsilon=1e-10
        cdef float data=0, coef=0, cdummy=0, cddummy=0
        cdef bint do_dummy=False, do_dark=False, do_flat=False, do_polarization=False, do_solidAngle=False
        cdef numpy.ndarray[numpy.float64_t, ndim = 1] outData = numpy.zeros(self.bins, dtype=numpy.float64)
        cdef numpy.ndarray[numpy.float64_t, ndim = 1] outCount = numpy.zeros(self.bins, dtype=numpy.float64)
        cdef numpy.ndarray[numpy.float32_t, ndim = 1] outMerge = numpy.zeros(self.bins, dtype=numpy.float32)
        cdef float[:] cdata, tdata, cflat, cdark, csolidAngle, cpolarization

        #Ugly hack against bug #89: https://github.com/pyFAI/pyFAI/issues/89
        cdef int rc_before, rc_after
        rc_before = sys.getrefcount(self._lut)
        cdef vector[ vector[lut_point] ] lut = self._lut
        rc_after = sys.getrefcount(self._lut)
        cdef bint need_decref = NEED_DECREF & ((rc_after-rc_before)>=2)


        assert size == weights.size

        if dummy is not None:
            do_dummy = True
            cdummy =  <float>float(dummy)
            if delta_dummy is None:
                cddummy = <float>0.0
            else:
                cddummy = <float>float(delta_dummy)

        if flat is not None:
            do_flat = True
            assert flat.size == size
            cflat = numpy.ascontiguousarray(flat.ravel(), dtype=numpy.float32)
        if dark is not None:
            do_dark = True
            assert dark.size == size
            cdark = numpy.ascontiguousarray(dark.ravel(), dtype=numpy.float32)
        if solidAngle is not None:
            do_solidAngle = True
            assert solidAngle.size == size
            csolidAngle = numpy.ascontiguousarray(solidAngle.ravel(), dtype=numpy.float32)
        if polarization is not None:
            do_polarization = True
            assert polarization.size == size
            cpolarization = numpy.ascontiguousarray(polarization.ravel(), dtype=numpy.float32)

        if (do_dark + do_flat + do_polarization + do_solidAngle):
            tdata = numpy.ascontiguousarray(weights.ravel(), dtype=numpy.float32)
            cdata = numpy.zeros(size,dtype=numpy.float32)
            if do_dummy:
                for i in prange(size, nogil=True, schedule="static"):
                    data = tdata[i]
                    if ((cddummy!=0) and (fabs(data-cdummy) > cddummy)) or ((cddummy==0) and (data!=cdummy)):
                        #Nota: -= and /= operatore are seen as reduction in cython parallel.
                        if do_dark:
                            data = data - cdark[i]
                        if do_flat:
                            data = data / cflat[i]
                        if do_polarization:
                            data = data / cpolarization[i]
                        if do_solidAngle:
                            data = data / csolidAngle[i]
                        cdata[i]+=data
                    else: #set all dummy_like values to cdummy. simplifies further processing
                        cdata[i]+=cdummy
            else:
                for i in prange(size, nogil=True, schedule="static"):
                    data = tdata[i]
                    if do_dark:
                        data = data - cdark[i]
                    if do_flat:
                        data = data / cflat[i]
                    if do_polarization:
                        data = data / cpolarization[i]
                    if do_solidAngle:
                        data = data / csolidAngle[i]
                    cdata[i]+=data
        else:
            if do_dummy:
                tdata = numpy.ascontiguousarray(weights.ravel(), dtype=numpy.float32)
                cdata = numpy.zeros(size,dtype=numpy.float32)
                for i in prange(size, nogil=True, schedule="static"):
                    data = tdata[i]
                    if ((cddummy!=0) and (fabs(data-cdummy) > cddummy)) or ((cddummy==0) and (data!=cdummy)):
                        cdata[i]+=data
                    else:
                        cdata[i]+=cdummy
            else:
                cdata = numpy.ascontiguousarray(weights.ravel(), dtype=numpy.float32)
        #TODO: what is the best: static or guided ?
        for i in prange(bins, nogil=True, schedule="guided"):
            sum_data = 0.0
            sum_count = 0.0
            for j in range(lut[i].size()):
                idx = lut[i][j].idx
                coef = lut[i][j].coef
                if idx <= 0 and coef <= 0.0:
                    continue
                data = cdata[idx]
                if do_dummy and data==cdummy:
                    continue

                sum_data = sum_data + coef * data
                sum_count = sum_count + coef
            outData[i] += sum_data
            outCount[i] += sum_count
            if sum_count > epsilon:
                outMerge[i] += sum_data / sum_count
            else:
                outMerge[i] += cdummy

        #Ugly against bug#89
        if need_decref and (sys.getrefcount(self._lut)>=rc_before+2):
            print("Decref needed")
            Py_XDECREF(<PyObject *> self._lut)
        return  self.outPos, outMerge, outData, outCount



################################################################################
# Bidimensionnal regrouping
################################################################################

class HistoBBox2d(object):
    """
    2D histogramming with pixel splitting based on a look-up table

    The initialization of the class can take quite a while (operation are not parallelized)
    but each integrate is parallelized and quite efficient.
    """
    @cython.boundscheck(False)
    def __init__(self,
                 pos0,
                 delta_pos0,
                 pos1,
                 delta_pos1,
                 bins=(100,36),
                 pos0Range=None,
                 pos1Range=None,
                 mask=None,
                 mask_checksum=None,
                 allow_pos0_neg=False,
                 unit="undefined",
                 chiDiscAtPi=True
                 ):
        """
        @param pos0: 1D array with pos0: tth or q_vect
        @param delta_pos0: 1D array with delta pos0: max center-corner distance
        @param pos1: 1D array with pos1: chi
        @param delta_pos1: 1D array with max pos1: max center-corner distance, unused !
        @param bins: number of output bins (tth=100, chi=36 by default)
        @param pos0Range: minimum and maximum  of the 2th range
        @param pos1Range: minimum and maximum  of the chi range
        @param mask: array (of int8) with masked pixels with 1 (0=not masked)
        @param allow_pos0_neg: enforce the q<0 is usually not possible
        @param chiDiscAtPi: boolean; by default the chi_range is in the range ]-pi,pi[ set to 0 to have the range ]0,2pi[
        @param unit: can be 2th_deg or r_nm^-1 ...
        """
        cdef numpy.int32_t i, size, bin0, bin1
        self.size = pos0.size
        assert delta_pos0.size == self.size
        assert pos1.size == self.size
        assert delta_pos1.size == self.size
        self.chiDiscAtPi = 1 if chiDiscAtPi else 0
        self.allow_pos0_neg =  allow_pos0_neg

        try:
            bins0, bins1 = tuple(bins)
        except:
            bins0 = bins1 = bins
        if bins0 <= 0:
            bins0 = 1
        if bins1 <= 0:
            bins1 = 1
        self.bins = (int(bins0),int(bins1))
        self.lut_size = 0
        if  mask is not None:
            assert mask.size == self.size
            self.check_mask = True
            self.cmask = numpy.ascontiguousarray(mask.ravel(), dtype=numpy.int8)
            if mask_checksum:
                self.mask_checksum = mask_checksum
            else:
                self.mask_checksum = crc32(mask)
        else:
            self.check_mask = False
            self.mask_checksum = None

        self.cpos0 = numpy.ascontiguousarray(pos0.ravel(), dtype=numpy.float32)
        self.dpos0 = numpy.ascontiguousarray(delta_pos0.ravel(), dtype=numpy.float32)
        self.cpos0_sup = numpy.empty_like(self.cpos0)
        self.cpos0_inf = numpy.empty_like(self.cpos0)
        self.pos0Range = pos0Range
        self.pos1Range = pos1Range

        self.cpos1 = numpy.ascontiguousarray((pos1).ravel(), dtype=numpy.float32)
        self.dpos1 = numpy.ascontiguousarray((delta_pos1).ravel(), dtype=numpy.float32)
        self.cpos1_sup = numpy.empty_like(self.cpos1)
        self.cpos1_inf = numpy.empty_like(self.cpos1)
        self.calc_boundaries(pos0Range, pos1Range)
        self.delta0 = (self.pos0_max - self.pos0_min) / float(bins0)
        self.delta1 = (self.pos1_max - self.pos1_min) / float(bins1)
        self.lut_max_idx = None
        self._lut = None
        self.calc_lut()
        self.outPos0 = numpy.linspace(self.pos0_min+0.5*self.delta0, self.pos0_maxin-0.5*self.delta0, bins0)
        self.outPos1 = numpy.linspace(self.pos1_min+0.5*self.delta1, self.pos1_maxin-0.5*self.delta1, bins1)
        self.unit=unit
        self._lut_checksum = None #Calculated at export time to python


    @cython.boundscheck(False)
    @cython.wraparound(False)
    def calc_boundaries(self, pos0Range, pos1Range):
        """
        Called by constructor to calculate the boundaries and the bin position
        """

        cdef numpy.int32_t size = self.cpos0.size
        cdef bint check_mask = self.check_mask
        cdef numpy.int8_t[:] cmask
        cdef float[:] cpos0, dpos0, cpos0_sup, cpos0_inf
        cdef float[:] cpos1, dpos1, cpos1_sup, cpos1_inf,
        cdef float upper0, lower0, pos0_max, pos0_min, c0, d0
        cdef float upper1, lower1, pos1_max, pos1_min, c1, d1
        cdef bint allow_pos0_neg=self.allow_pos0_neg
        cdef bint chiDiscAtPi = self.chiDiscAtPi

        cpos0_sup = self.cpos0_sup
        cpos0_inf = self.cpos0_inf
        cpos0 = self.cpos0
        dpos0 = self.dpos0
        cpos1_sup = self.cpos1_sup
        cpos1_inf = self.cpos1_inf
        cpos1 = self.cpos1
        dpos1 = self.dpos1
        pos0_min=cpos0[0]
        pos0_max=cpos0[0]
        pos1_min=cpos1[0]
        pos1_max=cpos1[0]

        if check_mask:
            cmask = self.cmask
        with nogil:
            for idx in range(size):
                c0 = cpos0[idx]
                d0 = dpos0[idx]
                lower0 = c0 - d0
                upper0 = c0 + d0
                c1 = cpos1[idx]
                d1 = dpos1[idx]
                lower1 = c1 - d1
                upper1 = c1 + d1
                if not allow_pos0_neg and lower0<0:
                    lower0=0
                if upper1 > (2-chiDiscAtPi)*pi:
                    upper1 = (2-chiDiscAtPi)*pi
                if lower1 < (-chiDiscAtPi)*pi:
                    lower1 = (-chiDiscAtPi)*pi
                cpos0_sup[idx] = upper0
                cpos0_inf[idx] = lower0
                cpos1_sup[idx] = upper1
                cpos1_inf[idx] = lower1
                if not (check_mask and cmask[idx]):
                    if upper0>pos0_max:
                        pos0_max = upper0
                    if lower0<pos0_min:
                        pos0_min = lower0
                    if upper1>pos1_max:
                        pos1_max = upper1
                    if lower1<pos1_min:
                        pos1_min = lower1

        if pos0Range is not None and len(pos0Range) > 1:
            self.pos0_min = min(pos0Range)
            self.pos0_maxin = max(pos0Range)
        else:
            self.pos0_min = pos0_min
            self.pos0_maxin = pos0_max


        if pos1Range is not None and len(pos1Range) > 1:
            self.pos1_min = min(pos1Range)
            self.pos1_maxin = max(pos1Range)
        else:
            self.pos1_min = pos1_min
            self.pos1_maxin = pos1_max

        if (not allow_pos0_neg) and self.pos0_min < 0:
            self.pos0_min = 0
        self.pos0_max = self.pos0_maxin * EPS32
        self.cpos0_sup = cpos0_sup
        self.cpos0_inf = cpos0_inf
        self.pos1_max = self.pos1_maxin * EPS32
        self.cpos1_sup = cpos1_sup
        self.cpos1_inf = cpos1_inf


    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def calc_lut(self):
        'calculate the max number of elements in the LUT and populate it'
        cdef float delta0=self.delta0, pos0_min=self.pos0_min, min0, max0, fbin0_min, fbin0_max
        cdef float delta1=self.delta1, pos1_min=self.pos1_min, min1, max1, fbin1_min, fbin1_max
        cdef numpy.int32_t bin0_min, bin0_max, bins0 = self.bins[0]
        cdef numpy.int32_t bin1_min, bin1_max, bins1 = self.bins[1]
        cdef numpy.int32_t k, idx, lut_size, i, j, size=self.size
        cdef bint check_mask
        cdef float[:] cpos0_sup = self.cpos0_sup
        cdef float[:] cpos0_inf = self.cpos0_inf
        cdef float[:] cpos1_inf = self.cpos1_inf
        cdef float[:] cpos1_sup = self.cpos1_sup
        cdef numpy.ndarray[numpy.int32_t, ndim = 2] outMax = numpy.zeros((bins0,bins1), dtype=numpy.int32)
#        cdef numpy.ndarray[lut_point, ndim = 3] lut
        cdef lut_point[:,:,:] lut
        cdef numpy.int8_t[:] cmask
        if self.check_mask:
            cmask = self.cmask
            check_mask = True
        else:
            check_mask = False

#NOGIL
        with nogil:
            for idx in range(size):
                if (check_mask) and (cmask[idx]):
                    continue

                min0 = cpos0_inf[idx]
                max0 = cpos0_sup[idx]
                min1 = cpos1_inf[idx]
                max1 = cpos1_sup[idx]

                bin0_min = < numpy.int32_t > get_bin_number(min0, pos0_min, delta0)
                bin0_max = < numpy.int32_t > get_bin_number(max0, pos0_min, delta0)

                bin1_min = < numpy.int32_t > get_bin_number(min1, pos1_min, delta1)
                bin1_max = < numpy.int32_t > get_bin_number(max1, pos1_min, delta1)

                if (bin0_max < 0) or (bin0_min >= bins0) or (bin1_max < 0) or (bin1_min >= bins1):
                    continue

                if bin0_max >= bins0 :
                    bin0_max = bins0 - 1
                if  bin0_min < 0:
                    bin0_min = 0
                if bin1_max >= bins1 :
                    bin1_max = bins1 - 1
                if  bin1_min < 0:
                    bin1_min = 0

                for i in range(bin0_min, bin0_max+1):
                    for j in range(bin1_min , bin1_max+1):
                        outMax[i, j] +=  1

        self.lut_size = lut_size = outMax.max()
        outMax[:, :] = 0

        lut_nbytes = bins0 * bins1 * lut_size * sizeof(lut_point)
        if (os.name == "posix") and ("SC_PAGE_SIZE" in os.sysconf_names) and ("SC_PHYS_PAGES" in os.sysconf_names):
            try:
                memsize = os.sysconf("SC_PAGE_SIZE") * os.sysconf("SC_PHYS_PAGES")
            except OSError:
                pass
                if memsize < lut_nbytes:
                    raise MemoryError("Lookup-table (%i, %i, %i) is %.3fGB whereas the memory of the system is only %s"%(bins0, bins1, lut_size, lut_nbytes, memsize))
        # else hope we have enough memory
        lut = view.array(shape=(bins0, bins1, lut_size), itemsize=sizeof(lut_point), format="if")
#        lut = numpy.recarray(shape=(bins0, bins1, lut_size),dtype=[("idx",numpy.int32),("coef",numpy.float32)])
        memset(&lut[0,0,0], 0, lut_nbytes)

        #NOGIL
        with nogil:
            for idx in range(size):
                if (check_mask) and cmask[idx]:
                    continue

                min0 = cpos0_inf[idx]
                max0 = cpos0_sup[idx]
                min1 = cpos1_inf[idx]
                max1 = cpos1_sup[idx]

                fbin0_min = get_bin_number(min0, pos0_min, delta0)
                fbin0_max = get_bin_number(max0, pos0_min, delta0)
                fbin1_min = get_bin_number(min1, pos1_min, delta1)
                fbin1_max = get_bin_number(max1, pos1_min, delta1)

                bin0_min = <int> fbin0_min
                bin0_max = <int> fbin0_max
                bin1_min = <int> fbin1_min
                bin1_max = <int> fbin1_max

                if (bin0_max < 0) or (bin0_min >= bins0) or (bin1_max < 0) or (bin1_min >= bins1):
                    continue

                if bin0_max >= bins0 :
                    bin0_max = bins0 - 1
                if  bin0_min < 0:
                    bin0_min = 0
                if bin1_max >= bins1 :
                    bin1_max = bins1 - 1
                if  bin1_min < 0:
                    bin1_min = 0

                if bin0_min == bin0_max:
                    if bin1_min == bin1_max:
                        #All pixel is within a single bin
                        k = outMax[bin0_min, bin1_min]
                        lut[bin0_min, bin1_min, k].idx = idx
                        lut[bin0_min, bin1_min, k].coef = 1.0
                        outMax[bin0_min, bin1_min]= k+1

                    else:
                        #spread on more than 2 bins
                        deltaD = (< float > (bin1_min + 1)) - fbin1_min
                        deltaU = fbin1_max - ( bin1_max)
                        deltaA = 1.0 / (fbin1_max - fbin1_min)

                        k = outMax[bin0_min, bin1_min]
                        lut[bin0_min, bin1_min, k].idx = idx
                        lut[bin0_min, bin1_min, k].coef =  deltaA * deltaD
                        outMax[bin0_min, bin1_min] += 1

                        k = outMax[bin0_min, bin1_max]
                        lut[bin0_min, bin1_max, k].idx = idx
                        lut[bin0_min, bin1_max, k].coef =  deltaA * deltaU
                        outMax[bin0_min, bin1_max] += 1

                        for j in range(bin1_min + 1, bin1_max):
                            k = outMax[bin0_min, j]
                            lut[bin0_min, j, k].idx = idx
                            lut[bin0_min, j, k].coef =  deltaA
                            outMax[bin0_min, j] += 1

                else: #spread on more than 2 bins in dim 0
                    if bin1_min == bin1_max:
                        #All pixel fall on 1 bins in dim 1
                        deltaA = 1.0 / (fbin0_max - fbin0_min)
                        deltaL = (< float > (bin0_min + 1)) - fbin0_min

                        k = outMax[bin0_min, bin1_min]
                        lut[bin0_min, bin1_min, k].idx = idx
                        lut[bin0_min, bin1_min, k].coef =  deltaA * deltaL
                        outMax[bin0_min, bin1_min] = k+1

                        deltaR = fbin0_max - (< float > bin0_max)

                        k = outMax[bin0_max, bin1_min]
                        lut[bin0_max, bin1_min, k].idx = idx
                        lut[bin0_max, bin1_min, k].coef =  deltaA * deltaR
                        outMax[bin0_max, bin1_min] += 1

                        for i in range(bin0_min + 1, bin0_max):
                            k = outMax[i, bin1_min]
                            lut[i, bin1_min ,k].idx = idx
                            lut[i, bin1_min, k].coef =  deltaA
                            outMax[i, bin1_min] += 1

                    else:
                        #spread on n pix in dim0 and m pixel in dim1:
                        deltaL = (< float > (bin0_min + 1)) - fbin0_min
                        deltaR = fbin0_max - (< float > bin0_max)
                        deltaD = (< float > (bin1_min + 1)) - fbin1_min
                        deltaU = fbin1_max - (< float > bin1_max)
                        deltaA = 1.0 / ((fbin0_max - fbin0_min) * (fbin1_max - fbin1_min))

                        k = outMax[bin0_min, bin1_min]
                        lut[bin0_min, bin1_min ,k].idx = idx
                        lut[bin0_min, bin1_min, k].coef =  deltaA * deltaL * deltaD
                        outMax[bin0_min, bin1_min] += 1

                        k = outMax[bin0_min, bin1_max]
                        lut[bin0_min, bin1_max, k].idx = idx
                        lut[bin0_min, bin1_max, k].coef =  deltaA * deltaL * deltaU
                        outMax[bin0_min, bin1_max] += 1

                        k = outMax[bin0_max, bin1_min]
                        lut[bin0_max, bin1_min, k].idx = idx
                        lut[bin0_max, bin1_min, k].coef =  deltaA * deltaR * deltaD
                        outMax[bin0_max, bin1_min] += 1

                        k = outMax[bin0_max, bin1_max]
                        lut[bin0_max, bin1_max, k].idx = idx
                        lut[bin0_max, bin1_max, k].coef =  deltaA * deltaR * deltaU
                        outMax[bin0_max, bin1_max] += 1

                        for i in range(bin0_min + 1, bin0_max):
                            k = outMax[i, bin1_min]
                            lut[i, bin1_min, k].idx = idx
                            lut[i, bin1_min, k].coef =  deltaA * deltaD
                            outMax[i, bin1_min] += 1

                            for j in range(bin1_min + 1, bin1_max):
                                k = outMax[i, j]
                                lut[i, j, k].idx = idx
                                lut[i, j, k].coef =  deltaA
                                outMax[i, j] += 1

                            k = outMax[i, bin1_max]
                            lut[i, bin1_max, k].idx = idx
                            lut[i, bin1_max, k].coef =  deltaA * deltaU
                            outMax[i, bin1_max] += 1

                        for j in range(bin1_min + 1, bin1_max):
                            k = outMax[bin0_min, j]
                            lut[bin0_min, j, k].idx = idx
                            lut[bin0_min, j, k].coef =  deltaA * deltaL
                            outMax[bin0_min, j] += 1

                            k = outMax[bin0_max, j]
                            lut[bin0_max, j, k].idx = idx
                            lut[bin0_max, j, k].coef =  deltaA * deltaR
                            outMax[bin0_max, j] += 1

        self.lut_max_idx = outMax
        self._lut = lut

    def get_lut(self):
        """Getter for the LUT as actual numpy array"""
        cdef int rc_before, rc_after
        rc_before = sys.getrefcount(self._lut)
        cdef lut_point[:,:,:] lut = self._lut
        rc_after = sys.getrefcount(self._lut)
        cdef bint need_decref = NEED_DECREF and ((rc_after-rc_before)>=2)
        shape = (self._lut.shape[0]*self._lut.shape[1], self._lut.shape[2])
        cdef numpy.ndarray[numpy.float64_t, ndim=2] tmp_ary = numpy.empty(shape=shape, dtype=numpy.float64)
        memcpy(&tmp_ary[0,0], &lut[0,0,0], self._lut.nbytes)
        self.lut_checksum = crc32(tmp_ary)

        #Ugly against bug#89
        if need_decref and (sys.getrefcount(self._lut)>=rc_before+2):
            print("Warning: Decref needed")
            Py_XDECREF(<PyObject *> self._lut)

#        return tmp_ary.view(dtype=dtype_lut)
        return numpy.core.records.array(tmp_ary.view(dtype=dtype_lut),
                                        shape=shape,dtype=dtype_lut,
                                        copy=True)

    lut = property(get_lut)

    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    def integrate(self, weights, dummy=None, delta_dummy=None, dark=None, flat=None, solidAngle=None, polarization=None):
        """
        Actually perform the 2D integration which in this case looks more like a matrix-vector product

        @param weights: input image
        @type weights: ndarray
        @param dummy: value for dead pixels (optional)
        @type dummy: float
        @param delta_dummy: precision for dead-pixel value in dynamic masking
        @type delta_dummy: float
        @param dark: array with the dark-current value to be subtracted (if any)
        @type dark: ndarray
        @param flat: array with the dark-current value to be divided by (if any)
        @type flat: ndarray
        @param solidAngle: array with the solid angle of each pixel to be divided by (if any)
        @type solidAngle: ndarray
        @param polarization: array with the polarization correction values to be divided by (if any)
        @type polarization: ndarray
        @return:  I(2d), edges0(1d), edges1(1d), weighted histogram(2d), unweighted histogram (2d)
        @rtype: 5-tuple of ndarrays

        """
        cdef numpy.int32_t i=0, j=0, idx=0, bins0=self.bins[0], bins1=self.bins[1], bins=bins0*bins1, lut_size=self.lut_size, size=self.size, i0=0, i1=0
        cdef double sum_data=0, sum_count=0, epsilon=1e-10
        cdef float data=0, coef=0, cdummy=0, cddummy=0
        cdef bint do_dummy=False, do_dark=False, do_flat=False, do_polarization=False, do_solidAngle=False
        cdef numpy.ndarray[numpy.float64_t, ndim = 2] outData = numpy.zeros(self.bins, dtype=numpy.float64)
        cdef numpy.ndarray[numpy.float64_t, ndim = 2] outCount = numpy.zeros(self.bins, dtype=numpy.float64)
        cdef numpy.ndarray[numpy.float32_t, ndim = 2] outMerge = numpy.zeros(self.bins, dtype=numpy.float32)

        #Ugly hack against bug #89
        cdef int rc_before, rc_after
        rc_before = sys.getrefcount(self._lut)
        cdef lut_point[:,:,:] lut = self._lut
        rc_after = sys.getrefcount(self._lut)
        cdef bint need_decref = NEED_DECREF and ((rc_after-rc_before)>=2)


        cdef float[:] cdata, tdata, cflat, cdark, csolidAngle, cpolarization

        assert size == weights.size

        if dummy is not None:
            do_dummy = True
            cdummy =  <float>float(dummy)
            if delta_dummy is None:
                cddummy = <float>0.0
            else:
                cddummy = <float>float(delta_dummy)

        if flat is not None:
            do_flat = True
            assert flat.size == size
            cflat = numpy.ascontiguousarray(flat.ravel(), dtype=numpy.float32)
        if dark is not None:
            do_dark = True
            assert dark.size == size
            cdark = numpy.ascontiguousarray(dark.ravel(), dtype=numpy.float32)
        if solidAngle is not None:
            do_solidAngle = True
            assert solidAngle.size == size
            csolidAngle = numpy.ascontiguousarray(solidAngle.ravel(), dtype=numpy.float32)
        if polarization is not None:
            do_polarization = True
            assert polarization.size == size
            cpolarization = numpy.ascontiguousarray(polarization.ravel(), dtype=numpy.float32)

        if (do_dark + do_flat + do_polarization + do_solidAngle):
            tdata = numpy.ascontiguousarray(weights.ravel(), dtype=numpy.float32)
            cdata = numpy.zeros(size,dtype=numpy.float32)
            if do_dummy:
                for i in prange(size, nogil=True, schedule="static"):
                    data = tdata[i]
                    if ((cddummy!=0) and (fabs(data-cdummy) > cddummy)) or ((cddummy==0) and (data!=cdummy)):
                        #Nota: -= and /= operatore are seen as reduction in cython parallel.
                        if do_dark:
                            data = data - cdark[i]
                        if do_flat:
                            data = data / cflat[i]
                        if do_polarization:
                            data = data / cpolarization[i]
                        if do_solidAngle:
                            data = data / csolidAngle[i]
                        cdata[i]+=data
                    else: #set all dummy_like values to cdummy. simplifies further processing
                        cdata[i]+=cdummy
            else:
                for i in prange(size, nogil=True, schedule="static"):
                    data = tdata[i]
                    if do_dark:
                        data = data - cdark[i]
                    if do_flat:
                        data = data / cflat[i]
                    if do_polarization:
                        data = data / cpolarization[i]
                    if do_solidAngle:
                        data = data / csolidAngle[i]
                    cdata[i]+=data
        else:
            if do_dummy:
                tdata = numpy.ascontiguousarray(weights.ravel(), dtype=numpy.float32)
                cdata = numpy.zeros(size,dtype=numpy.float32)
                for i in prange(size, nogil=True, schedule="static"):
                    data = tdata[i]
                    if ((cddummy!=0) and (fabs(data-cdummy) > cddummy)) or ((cddummy==0) and (data!=cdummy)):
                        cdata[i]+=data
                    else:
                        cdata[i]+=cdummy
            else:
                cdata = numpy.ascontiguousarray(weights.ravel(), dtype=numpy.float32)
        #TODO: what is the best: static or guided ?
        for i0 in prange(bins0, nogil=True, schedule="guided"):
            for i1 in range(bins1):
                sum_data = 0.0
                sum_count = 0.0
                for j in range(lut_size):
                    idx = lut[i0, i1, j].idx
                    coef = lut[i0, i1, j].coef
                    if idx <= 0 and coef <= 0.0:
                        continue
                    data = cdata[idx]
                    if do_dummy and data==cdummy:
                        continue

                    sum_data = sum_data + coef * data
                    sum_count = sum_count + coef
                outData[i0, i1] += sum_data
                outCount[i0, i1] += sum_count
                if sum_count > epsilon:
                    outMerge[i0, i1] += sum_data / sum_count
                else:
                    outMerge[i0, i1] += cdummy

        #Ugly against bug #89
        if need_decref and (sys.getrefcount(self._lut)>=rc_before+2):
            Py_XDECREF(<PyObject *> self._lut)
        return  outMerge.T, self.outPos0, self.outPos1, outData.T, outCount.T




#!/usr/bin/python
# coding: utf-8
import sys, scipy
import matplotlib
matplotlib.use('Qt4Agg')
import pylab
from math import sqrt
import fabio,numpy
from utilstest import UtilsTest, getLogger
logger = getLogger(__file__)
pyFAI = sys.modules["pyFAI"]
from pyFAI.azimuthalIntegrator import AzimuthalIntegrator
from pyFAI.blob_detection import BlobDetection
from pyFAI.detectors import detector_factory


def somme(im):
    im[1:-1, 1:-1] += im[:-2, 1:-1] + im[2:, 1:-1] + im[1:-1, :-2] + im[1:-1, 2:] #+ im[:-2, :-2] + im[2:, 2:] + im[2:, :-2] + im[:-2, 2:]
    return im

def image_test():
    "Creating a test image containing several gaussian of several sizes"
    img = numpy.zeros((128*4,128*4))
    a = numpy.linspace(0.5, 8, 16)
    xc = [64,64,64,64,192,192,192,192,320,320,320,320,448,448,448,448]
    yc = [64,192,320,448,64,192,320,448,64,192,320,448,64,192,320,448]
    cpt = 0
    for sigma in a:
        img = make_gaussian(img,sigma,xc[cpt],yc[cpt])
        cpt = cpt + 1
    img = add_noise(img, 0.1)
    return img

def image_test_rings():
    "Creating a test image containing gaussian spots on several rings"
    rings = 10
    mod = 50
    detector = detector_factory("Titan")
    sigma = detector.pixel1 * 4
    shape = detector.max_shape
    ai = AzimuthalIntegrator(detector=detector)
    ai.setFit2D(1000, 1000, 1000)
    r = ai.rArray(shape)
    r_max = r.max()
    chi = ai.chiArray(shape)
    img = numpy.zeros(shape)
    modulation = (1 + numpy.sin(5 * r + chi * mod))
    for radius in numpy.linspace(0, r_max, rings):
        img += numpy.exp(-(r - radius) ** 2 / (2 * (sigma * sigma)))
    
    img *= modulation
    img = add_noise(img, 0.0)
    return img

def add_noise(img,rate):
    noise = numpy.random.random(img.shape) * rate
    return img+noise
    
def make_gaussian(im,sigma,xc,yc):
    "Creating 2D gaussian to be put in a test image"
    e = 1
    angle = 0
    sx = sigma * (1+e)
    sy = sigma * (1-e)
    size = int( 8*sigma +1 )
    if size%2 == 0 :
        size += 1
    x = numpy.arange(0, size, 1, float)
    y = x[:,numpy.newaxis]
#     x = x * 2
    x0 = y0 = size // 2
    gausx = numpy.exp(-4*numpy.log(2) * (x-x0)**2 / sx**2)
    gausy = numpy.exp(-4*numpy.log(2) * (y-y0)**2 / sy**2)
    gaus = 0.01 + gausx * gausy
    im[xc-size/2:xc+size/2+1,yc-size/2:yc+size/2+1] = scipy.ndimage.rotate(gaus,angle, reshape = False)
    return im


if len(UtilsTest.options.args) > 0:
     data = fabio.open(UtilsTest.options.args[0]).data
     if len(UtilsTest.options.args) > 1:
         msk = fabio.open(UtilsTest.options.args[1]).data
     else:
         msk = None
else:
    data = image_test_rings()
    msk = None


bd = BlobDetection(data, mask=msk)#, cur_sigma=0.25, init_sigma=numpy.sqrt(2)/2, dest_sigma=numpy.sqrt(2), scale_per_octave=2)

pylab.ion()
f=pylab.figure(1)
ax = f.add_subplot(111)
ax.imshow(numpy.log1p(data), interpolation = 'nearest')

for i in range(5):
    print ('Octave #%i' %i)
    bd._one_octave(shrink=True, refine = True, n_5 = False)
    print("Octave #%i Total kp: %i" % (i, bd.keypoints.size))
    
# bd._one_octave(False, True ,False)
    
print ('Final size of keypoints : %i'% bd.keypoints.size)

i = 0
# for kp  in bd.keypoints:
#     ds = kp.sigma
#     ax.annotate("", xy=(kp.x, kp.y), xytext=(kp.x+ds, kp.y+ds),
#                 arrowprops=dict(facecolor='blue', shrink=0.05),)
sigma = bd.keypoints.sigma
for i,c in enumerate("ygrcmykw"):
#    j = 2 ** i
    m = numpy.logical_and(sigma >= i, sigma < (i + 1))
    ax.plot(bd.keypoints[m].x, bd.keypoints[m].y, "o" + c, label=str(i))
ax.legend()

if sigma.size > 0:
    h = pylab.figure(2)
    x, y, o = pylab.hist(sigma, bins=100)
    h.show()
      
    index = numpy.where(x == x.max())
    kp = bd.keypoints[bd.keypoints.sigma > y[index]]
else : kp = bd.keypoints

# pylab.figure()
# pylab.imshow(numpy.log1p(data), interpolation = 'nearest')
# pylab.plot(kp.x,kp.y,'og')

f.show()
raw_input()



# -*- coding: utf-8 -*-
# Copyright (C) 2012, Almar Klein
# Copyright (C) 2014, Jerome Kieffer (European Synchrotron Radiation Facility)
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the <organization> nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Cythonized version of the marching square function for "isocontour" plot
"""
__authors__ = ["Almar Klein", "Jerome Kieffer"]
__contact__ = "Jerome.kieffer@esrf.fr"
__date__ = "20/10/2014"
__status__ = "stable"
__license__ = "BSD-3 clauses"

# Cython specific imports

import numpy 
cimport numpy
import cython
from libc.math cimport M_PI, sin, floor, fabs
cdef double epsilon = numpy.finfo(numpy.float64).eps
from cython.view cimport array as cvarray
from .utils import timeit

cdef numpy.int8_t[:, :] EDGETORELATIVEPOSX = numpy.array([[0, 1], [1, 1], [1, 0], [0, 0], [0, 1], [1, 1], [1, 0], [0, 0], [0, 0], [1, 1], [1, 1], [0, 0]], dtype=numpy.int8)
cdef numpy.int8_t[:, :] EDGETORELATIVEPOSY = numpy.array([[0, 0], [0, 1], [1, 1], [1, 0], [0, 0], [0, 1], [1, 1], [1, 0], [0, 0], [0, 0], [1, 1], [1, 1]], dtype=numpy.int8)
cdef numpy.int8_t[:, :] EDGETORELATIVEPOSZ = numpy.array([[0, 0], [0, 0], [0, 0], [0, 0], [1, 1], [1, 1], [1, 1], [1, 1], [0, 1], [0, 1], [0, 1], [0, 1]], dtype=numpy.int8)
cdef numpy.int8_t[:, :] CELLTOEDGE = numpy.array([
                                                    [0, 0, 0, 0, 0],  # Case 0: nothing
                                                    [1, 0, 3, 0, 0],  # Case 1
                                                    [1, 0, 1, 0, 0],  # Case 2
                                                    [1, 1, 3, 0, 0],  # Case 3
                             
                                                    [1, 1, 2, 0, 0],  # Case 4
                                                    [2, 0, 1, 2, 3],  # Case 5 > ambiguous
                                                    [1, 0, 2, 0, 0],  # Case 6
                                                    [1, 2, 3, 0, 0],  # Case 7
                             
                                                    [1, 2, 3, 0, 0],  # Case 8
                                                    [1, 0, 2, 0, 0],  # Case 9
                                                    [2, 0, 3, 1, 2],  # Case 10 > ambiguous
                                                    [1, 1, 2, 0, 0],  # Case 11
                            
                                                    [1, 1, 3, 0, 0],  # Case 12
                                                    [1, 0, 1, 0, 0],  # Case 13
                                                    [1, 0, 3, 0, 0],  # Case 14
                                                    [0, 0, 0, 0, 0],  # Case 15
                                                ], dtype=numpy.int8)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def marching_squares(float[:, :] img, double isovalue,
                     numpy.int8_t[:, :] cellToEdge, 
                     numpy.int8_t[:, :] edgeToRelativePosX, 
                     numpy.int8_t[:, :] edgeToRelativePosY):
    cdef:
        int dim_y = img.shape[0]
        int dim_x = img.shape[1]
        numpy.ndarray[numpy.float32_t, ndim = 2] edges = numpy.zeros((dim_x * dim_y, 2), numpy.float32)
        int x, y, z, i, j, k, index, edgeCount = 0
        int dx1, dy1, dz1, dx2, dy2, dz2
        double fx, fy, fz, ff, tmpf, tmpf1, tmpf2
    with nogil:
        for y in range(dim_y - 1):
            for x in range(dim_x - 1):
    
                # Calculate index.
                index = 0
                if img[y, x] > isovalue:
                    index += 1
                if img[y, x + 1] > isovalue:
                    index += 2
                if img[y + 1, x + 1] > isovalue:
                    index += 4
                if img[y + 1, x] > isovalue:
                    index += 8
    
                # Resolve ambiguity
                if index == 5 or index == 10:
                    # Calculate value of cell center (i.e. average of corners)
                    tmpf = 0.25*(img[y,x]+img[y,x+1]+img[y+1,x]+img[y+1,x+1])
                    # If below isovalue, swap
                    if tmpf <= isovalue:
                        if index == 5:
                            index = 10
                        else:
                            index = 5
    
                # For each edge ...
                for i in range(cellToEdge[index,0]):
                    # For both ends of the edge ...
                    for j in range(2):
                        # Get edge index
                        k = cellToEdge[index, 1+i*2+j]
                        # Use these to look up the relative positions of the pixels to interpolate
                        dx1, dy1 = edgeToRelativePosX[k,0], edgeToRelativePosY[k,0]
                        dx2, dy2 = edgeToRelativePosX[k,1], edgeToRelativePosY[k,1]
                        # Define "strength" of each corner of the cube that we need
                        tmpf1 = 1.0 / (epsilon + fabs( img[y+dy1,x+dx1] - isovalue))
                        tmpf2 = 1.0 / (epsilon + fabs( img[y+dy2,x+dx2] - isovalue))
                        # Apply a kind of center-of-mass method
                        fx, fy, ff = 0.0, 0.0, 0.0
                        fx += <double> dx1 * tmpf1; fy += <double>dy1 * tmpf1;  ff += tmpf1
                        fx += <double> dx2 * tmpf2; fy += <double>dy2 * tmpf2;  ff += tmpf2
                        #
                        fx /= ff
                        fy /= ff
                        # Append point
                        edges[edgeCount,0] = <float> (x + fx)
                        edges[edgeCount,1] = <float> (y + fy)
                        edgeCount += 1

    # Done
    return edges[:edgeCount, : ]


@cython.boundscheck(False)
def sort_edges(edges):
    """
    Reorder edges in such a way they become contiguous
    """
    cdef:
        int size = edges.shape[0]
        int[:] pos = cvarray(shape=(size,), itemsize=sizeof(int), format="i")
        int[:] remaining = cvarray(shape=(size,), itemsize=sizeof(int), format="i")
        float[:, :] dist2 = cvarray(shape=(size, size), itemsize=sizeof(float), format="f")
        float d
        int i, j, index = 0, current = 0
        float[:, :]edges_ = numpy.ascontiguousarray(edges, numpy.float32)      
    dist2[:, :] = 0
    pos[:] = 0
    
    with nogil:
        # initialize the distance (squared) array:
        for i in range(size):
            remaining[i] = i
            for j in range(i + 1, size):
                d = (edges_[i, 0] - edges_[j, 0]) ** 2 + (edges_[i, 1] - edges_[j, 1]) ** 2 
                dist2[i, j] = d
                dist2[j, i] = d
        # set element in remaining to -1 when already transfered
        # O(n^2) implementation, not bright, any better idea is welcome
        remaining[0] = -1
        pos[0] = 0
        for i in range(1, size):
            current = pos[i - 1]
            index = -1
            for j in range(1, size):
                if remaining[j] == -1:
                    continue
                elif index == -1:  # not yet found a candidate
                    index = remaining[j]
                    d = dist2[index, current]
                    continue
                elif dist2[j, current] < d:    
                    index = j
                    d = dist2[current, index]
            pos[i] = index
            remaining[index] = -1
    return edges[pos, :]


def isocontour(img, isovalue=None, sorted=False):
    """ isocontour(img, isovalue=None)

    Calculate the iso contours for the given 2D image. If isovalue
    is not given or None, a value between the min and max of the image
    is used.

    @param img: 2D array representing the image
    @param isovalue: the value for which the iso_contour shall be calculated
    @param sorted: perform a sorting of the points to have them contiguous ? 

    Returns a pointset in which each two subsequent points form a line
    piece. This van be best visualized using "vv.plot(result, ls='+')".

    """

    # Check image
    if not isinstance(img, numpy.ndarray) or (img.ndim != 2):
        raise ValueError('img should be a 2D numpy array.')

    # Get isovalue
    if isovalue is None:
        isovalue = 0.5 * (img.min() + img.max())
    else:
        # Will raise error if not float-like value given
        isovalue = float(isovalue)
    res = marching_squares(numpy.ascontiguousarray(img, numpy.float32), isovalue,
                           CELLTOEDGE, EDGETORELATIVEPOSX, EDGETORELATIVEPOSY)
    if sorted:
        return sort_edges(res)
    else:
        return res



#!/usr/bin/python

import pyFAI
import numpy
from pyFAI import splitPixelFull, splitPixel

ai = pyFAI.AzimuthalIntegrator(detector="Fairchild")
shape = (2048, 2048)
data = numpy.zeros(shape)
data[100, 200] = 1

tth, I = ai.integrate1d(data, 10000, correctSolidAngle=False, method="splitpixel", unit="2th_deg")

res_splitPixelFull = splitPixelFull.fullSplit1D(ai._corner4Da, data, bins=10000)
res_splitPixel = splitPixel.fullSplit1D(ai._corner4Da, data, bins=10000)

for i, ary in enumerate(("tth", "I", "unweight", "weight")):
    print("Error on %s: %s" % (i, abs(res_splitPixelFull[i] - res_splitPixel[i]).max()))





#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#    Project: Fast Azimuthal Integration
#             https://github.com/pyFAI/pyFAI
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"Dummy test to run first to check for relative imports"

__author__ = "Jérôme Kieffer"
__contact__ = "Jerome.Kieffer@ESRF.eu"
__license__ = "GPLv3+"
__copyright__ = "European Synchrotron Radiation Facility, Grenoble, France"
__date__ = "15/12/2014"


import unittest
import sys
if __name__ == '__main__':
    import pkgutil, os
    __path__ = pkgutil.extend_path([os.path.dirname(__file__)], "pyFAI.test")
from .utilstest import UtilsTest, Rwp, getLogger
logger = getLogger(__file__)
pyFAI = sys.modules["pyFAI"]


class TestDummy(unittest.TestCase):
    def test_dummy(self):
        print(__name__)
        print(pyFAI)


def test_suite_all_dummy():
    testSuite = unittest.TestSuite()
    testSuite.addTest(TestDummy("test_dummy"))
    return testSuite

if __name__ == '__main__':
    mysuite = test_suite_all_dummy()
    runner = unittest.TextTestRunner()
    runner.run(mysuite)
    UtilsTest.clean_up()
